import math
import os
import re

import FreeCAD
import numpy as np
from numpy import linalg as LA

from .Objects import (
    Box,
    CadCell,
    Cone,
    Cylinder,
    Ellipsoid,
    EllipticCone,
    EllipticCylinder,
    HyperbolicCylinder,
    Hyperboloid,
    Paraboloid,
    Plane,
    Sphere,
    Torus,
)
from .Parser import parser as mp
from .remh import CellCardString, remove_hash


# TODO rename as there are two classes with this name
class McnpInput:
    def __init__(self, name):
        if not os.path.isfile(name):
            raise FileNotFoundError(f"File {name} does not exist")
        self.__inputcards__ = list(mp.get_cards(name))
        self.Transformations = self.__getTransList__()
        return

    def GetFilteredCells(self, Surfaces, config):
        levels, contLevels, Universes = self.GetLevelStructure()

        FilteredCells = {}

        Ustart = config["Ustart"]
        subUniverses = getSubUniverses(Ustart, Universes)
        subUniverses.add(Ustart)

        if config["levelMax"] == "all":
            levelMax = len(levels)
        else:
            levelMax = config["levelMax"] + 1

        levelUniverse = set()
        for lev in range(0, levelMax):
            for U in levels[lev]:
                levelUniverse.add(U)
        subUniverses = subUniverses.intersection(levelUniverse)

        for U in list(Universes.keys()):
            if U not in subUniverses:
                del Universes[U]

        for U in Universes.keys():
            FilteredCells[U] = selectCells(Universes[U], config)
            processSurfaces(FilteredCells[U], Surfaces)

        # change the surface name in surface dict
        newSurfaces = {}
        for k in Surfaces.keys():
            newkey = Surfaces[k].id
            newSurfaces[newkey] = Surfaces[k]

        for U, universe in FilteredCells.items():
            substituteLikeCell(universe, newSurfaces)

            # set cell as CAD cell Object
            for cname, c in universe.items():
                # print(cname,c.geom.str)
                universe[cname] = CadCell(c)

        return levels, FilteredCells, newSurfaces

    def GetLevelStructure(self):
        containers = []
        Universe_dict = {}
        cellCards = {}

        for c in self.__inputcards__:
            if c.ctype != mp.CID.cell:
                continue
            c.get_values()
            cstr = CellCardString("".join(c.lines))

            if cstr.TRCL:
                cstr.TRCL = TransformationMatrix(cstr.TRCL, self.Transformations)
            if cstr.TR:
                cstr.TR = TransformationMatrix(cstr.TR, self.Transformations)
            cellCards[c.name] = cstr

        setExplicitCellDefinition(cellCards)
        for cname, c in cellCards.items():
            if c.U is None:
                c.U = 0
            if c.U not in Universe_dict.keys():
                Universe_dict[c.U] = {}
            Universe_dict[c.U].update({c.name: c})

            if c.FILL:
                containers.append(c)

        currentLevel = [0]
        nextLevel = []
        contLevel = {0: [(0, 0)]}
        univLevel = {0: {0}}
        level = 0

        while True:
            level += 1
            contLevel[level] = []
            univLevel[level] = set()
            for c in reversed(containers):
                if c.U in currentLevel:
                    c.Level = level
                    nextLevel.append(c.FILL)
                    contLevel[level].append((c.U, c.name))
                    univLevel[level].add(c.FILL)

                    containers.remove(c)

            if nextLevel == []:
                break
            currentLevel = nextLevel
            nextLevel = []

        return univLevel, contLevel, Universe_dict

    def GetCells(self, U=None, Fill=None):
        cell_cards = {}
        for c in self.__inputcards__:
            if c.ctype != mp.CID.cell:
                continue
            c.get_values()
            U_cell = c.get_u()
            Fill_cell = c.get_f()
            if U is None and Fill is None:
                cell_cards[c.name] = c
            elif U_cell == U and U is not None:
                cell_cards[c.name] = c
            elif Fill_cell == Fill and Fill is not None:
                cell_cards[c.name] = c

        return cell_cards

    def GetSurfaces(self, scale=1.0):
        surf_cards = {}
        number = 1
        for c in self.__inputcards__:
            if c.ctype != mp.CID.surface:
                continue
            c.get_values()
            c.TR = TransformationMatrix(c.TR, self.Transformations)
            surf_cards[c.name] = (c.stype, c.scoefs, c.TR, number)
            number += 1

        # return surface as surface Objects type
        return Get_primitive_surfaces(surf_cards, scale)

    def __getTransList__(self):
        trl = {}
        for c in self.__inputcards__:
            if c.ctype != mp.CID.data:
                continue
            c.get_values()
            if c.dtype == "TRn":
                trValues = []
                for v in c.values[1:]:
                    trValues.append(v[0])
                trl[c.name] = getTransMatrix(trValues, c.unit)
        return trl


# fmt: off
def getTransMatrix(trsf, unit="", scale=10.0):

    if len(trsf) == 3:
        trsfMat = FreeCAD.Matrix(
            1, 0, 0, trsf[0] * scale,
            0, 1, 0, trsf[1] * scale,
            0, 0, 1, trsf[2] * scale,
            0, 0, 0, 1,
        )
    else:
        if unit == "*":
            coeff = tuple(map(math.radians, trsf[3:12]))
            coeff = tuple(map(math.cos, coeff))
        else:
            coeff = trsf[3:12]

        trsfMat = FreeCAD.Matrix(
            coeff[0], coeff[3], coeff[6], trsf[0] * scale,
            coeff[1], coeff[4], coeff[7], trsf[1] * scale,
            coeff[2], coeff[5], coeff[8], trsf[2] * scale,
            0, 0, 0, 1,
        )
    return trsfMat    
# fmt: on


def substituteLikeCell(universe, Surfaces):
    number = re.compile(r"\#?\s*\d+")
    newId = len(Surfaces)

    # create new cells objects for like cell cells
    for c in universe.values():
        if c.likeCell:
            c.geom = universe[c.likeCell].geom.copy()
        if not c.TRCL:
            continue  # likebut cell should have TRCL card

    # transform change cell the parameters if needed
    for c in universe.values():
        if not c.TRCL:
            continue
        cellSurf = c.geom.get_surfaces_numbers()
        surfDict = {}
        for surf in cellSurf:
            newId += 1
            surfDict[surf] = newId
            Surfaces[newId] = Surfaces[surf].copy()
            Surfaces[newId].id = newId
            Surfaces[newId].transform(c.TRCL)
        if c.FILL:
            c.TR = c.TRCL
        c.TRCL = None

        # rename the surface of the like cell
        pos = 0
        while True:
            m = number.search(c.geom.str, pos)
            if not m:
                break
            if "#" in m.group():
                pos = m.end()
                continue
            surf = int(m.group())
            pos = c.geom.replace(surf, surfDict[surf], pos)
            if pos < 0:
                break


def selectCells(cellList, config):
    selected = {}
    # options are 'all' material
    if config["mat"][0] == "all":
        if config["cell"][0] == "all":
            selected = cellList
        elif config["cell"][0] == "exclude":
            for name, c in cellList.items():
                if name not in config["cell"][1]:
                    selected[name] = c
        elif config["cell"][0] == "include":
            for name, c in cellList.items():
                if name in config["cell"][1]:
                    selected[name] = c

    # options are 'exclude' material
    elif config["mat"][0] == "exclude":
        if config["cell"][0] == "all":
            for name, c in cellList.items():
                if c.FILL is None:
                    if c.MAT not in config["mat"][1]:
                        selected[name] = c
                else:
                    selected[name] = c  # Fill cell are not tested against material number
        elif config["cell"][0] == "exclude":
            for name, c in cellList.items():
                if c.FILL is None:
                    if c.MAT not in config["mat"][1]:
                        if name not in config["cell"][1]:
                            selected[name] = c
                else:
                    if name not in config["cell"][1]:
                        selected[name] = c  # Fill cell are not tested against material number
        elif config["cell"][0] == "include":
            for name, c in cellList.items():
                if c.FILL is None:
                    if c.MAT not in config["mat"][1]:
                        if name in config["cell"][1]:
                            selected[name] = c
                else:
                    if name in config["cell"][1]:
                        selected[name] = c  # Fill cell are not tested against material number

    # options are 'include' material
    elif config["mat"][0] == "include":
        if config["cell"][0] == "all":
            for name, c in cellList.items():
                if c.FILL is None:
                    if c.MAT in config["mat"][1]:
                        selected[name] = c
                else:
                    selected[name] = c  # Fill cell are not tested against material number
        elif config["cell"][0] == "exclude":
            for name, c in cellList.items():
                if c.FILL is None:
                    if c.MAT in config["mat"][1]:
                        if name not in config["cell"][1]:
                            selected[name] = c
                else:
                    if name not in config["cell"][1]:
                        selected[name] = c  # Fill cell are not tested against material number
        elif config["cell"][0] == "include":
            for name, c in cellList.items():
                if c.FILL is None:
                    if c.MAT in config["mat"][1]:
                        if name in config["cell"][1]:
                            selected[name] = c
                else:
                    if name in config["cell"][1]:
                        selected[name] = c  # Fill cell are not tested against material number

    # remove complementary in cell of the universe
    for cname, c in selected.items():
        c.geom = remove_hash(cellList, cname)

    if not selected:
        raise ValueError("No cells selected. Check input or selection criteria in config file.")

    return selected


# Change implicit cell definition (like-but type or cell with #)
# to explicit cell definition (only surface number)
def setExplicitCellDefinition(universeCells):
    for cname, c in universeCells.items():
        if c.likeCell:
            lkc = universeCells[c.likeCell]
            c.geom = lkc.geom
            if not c.U:
                c.U = lkc.U
            if not c.FILL:
                c.FILL = lkc.FILL
            if not c.MAT:
                c.MAT = lkc.MAT
            if not c.TR:
                c.TR = lkc.TR
            if not c.TRCL:
                c.TRCL = lkc.TRCL
    return


def processSurfaces(UCells, Surfaces):
    number = re.compile(r"\#?\s*\d+")

    for cname, c in UCells.items():
        c.geom.remove_comments(full=True)
        pos = 0
        while True:
            m = number.search(c.geom.str, pos)
            if not m:
                break
            if "#" in m.group():
                pos = m.end()
                continue
            surf = int(m.group())
            if surf == 0:
                print(c.name)
                print(m)
                print(c.geom.str)
            pos = c.geom.replace(surf, Surfaces[surf].id, pos)


def getTransMatrix(trsf, unit="", scale=10.0):

    if len(trsf) == 3:
        trsfMat = FreeCAD.Matrix(
            1,
            0,
            0,
            trsf[0] * scale,
            0,
            1,
            0,
            trsf[1] * scale,
            0,
            0,
            1,
            trsf[2] * scale,
            0,
            0,
            0,
            1,
        )
    else:
        if unit == "*":
            coeff = tuple(map(math.radians, trsf[3:12]))
            coeff = tuple(map(math.cos, coeff))
        else:
            coeff = trsf[3:12]

        trsfMat = FreeCAD.Matrix(
            coeff[0],
            coeff[3],
            coeff[6],
            trsf[0] * scale,
            coeff[1],
            coeff[4],
            coeff[7],
            trsf[1] * scale,
            coeff[2],
            coeff[5],
            coeff[8],
            trsf[2] * scale,
            0,
            0,
            0,
            1,
        )
    return trsfMat


def TransformationMatrix(TRSF, Transformations):

    if TRSF:
        if type(TRSF) is int:
            return Transformations[TRSF]

        else:
            return getTransMatrix(TRSF)
    else:
        return TRSF


def getSubUniverses(Ustart, Universes):
    Uid = set()
    for c in Universes[Ustart].values():
        if c.FILL:
            Uid.add(c.FILL)

    AllU = Uid.copy()
    for U in Uid:
        AllU = AllU.union(getSubUniverses(U, Universes))

    return AllU


# traduce mcnp surface definition for Solid_Cell class
#  planes:
#     Stype = 'plane'
#     params = [ax,ay,az,d]
#
#  spheres:
#     Stype = 'shpere'
#     params = [cx,cy,cz,R]
#
#  cylinders:
#     Stype = 'cylinder'
#     params = [[px,py,pz],[vx,vy,vz],R]
#
#  cones:
#     Stype = 'cone'
#     params = [[px,py,pz],[vx,vy,vz],t,sht]
#
#  torus:
#     Stype = 'torus'
#     params = [[px,py,pz],[vx,vy,vz],ra,r]


# Return a diccionary with the corresponding surface Object
def Get_primitive_surfaces(mcnp_surfaces, scale=10.0):

    X_vec = FreeCAD.Vector(1.0, 0.0, 0.0)
    Y_vec = FreeCAD.Vector(0.0, 1.0, 0.0)
    Z_vec = FreeCAD.Vector(0.0, 0.0, 1.0)
    negX_vec = -X_vec
    negY_vec = -Y_vec
    negZ_vec = -Z_vec
    origin = FreeCAD.Vector(0.0, 0.0, 0.0)

    surfaces = {}
    for Sid in mcnp_surfaces.keys():
        MCNPtype = mcnp_surfaces[Sid][0].upper()
        MCNPparams = mcnp_surfaces[Sid][1]
        trsf = mcnp_surfaces[Sid][2]
        number = mcnp_surfaces[Sid][3]

        params = []
        Stype = None
        if MCNPtype in ("P", "PX", "PY", "PZ"):
            Stype = "plane"
            if MCNPtype == "P":
                if len(MCNPparams) == 4:
                    normal = FreeCAD.Vector(MCNPparams[0:3])
                    params = (normal, MCNPparams[3] * scale)
                else:
                    coeffs = points_to_coeffs(MCNPparams[0:9])
                    normal = FreeCAD.Vector(coeffs[0:3])
                    point = coeffs[3] / normal.Length
                    normal.normalize()
                    params = (normal, point * scale)
            elif MCNPtype == "PX":
                params = (X_vec, MCNPparams[0] * scale)
            elif MCNPtype == "PY":
                params = (Y_vec, MCNPparams[0] * scale)
            elif MCNPtype == "PZ":
                params = (Z_vec, MCNPparams[0] * scale)

        elif MCNPtype in ["S", "SX", "SY", "SZ", "SO", "SPH"]:
            Stype = "sphere"
            if MCNPtype in ["S", "SPH"]:
                params = (
                    FreeCAD.Vector(MCNPparams[0:3]) * scale,
                    MCNPparams[3] * scale,
                )
            elif MCNPtype == "SX":
                params = (
                    FreeCAD.Vector(MCNPparams[0] * scale, 0.0, 0.0),
                    MCNPparams[1] * scale,
                )
            elif MCNPtype == "SY":
                params = (
                    FreeCAD.Vector(0.0, MCNPparams[0] * scale, 0.0),
                    MCNPparams[1] * scale,
                )
            elif MCNPtype == "SZ":
                params = (
                    FreeCAD.Vector(0.0, 0.0, MCNPparams[0] * scale),
                    MCNPparams[1] * scale,
                )
            elif MCNPtype == "SO":
                params = (origin, MCNPparams[0] * scale)

        elif MCNPtype in ["CX", "C/X", "CY", "C/Y", "CZ", "C/Z"]:
            if MCNPtype in ["CX", "CY", "CZ"]:
                R = MCNPparams[0]
            else:
                R = MCNPparams[2]
                x1 = MCNPparams[0]
                x2 = MCNPparams[1]

            Stype = "cylinder"

            if MCNPtype == "CX":
                v = X_vec
                p = origin
            elif MCNPtype == "CY":
                v = Y_vec
                p = origin
            elif MCNPtype == "CZ":
                v = Z_vec
                p = origin
            elif MCNPtype == "C/X":
                v = X_vec
                p = FreeCAD.Vector(0.0, x1, x2)
            elif MCNPtype == "C/Y":
                v = Y_vec
                p = FreeCAD.Vector(x1, 0.0, x2)
            elif MCNPtype == "C/Z":
                v = Z_vec
                p = FreeCAD.Vector(x1, x2, 0.0)

            if scale != 1.0:
                p = p.multiply(scale)
                R *= scale

            params = (p, v, R)

        elif MCNPtype in ["KX", "KY", "KZ"]:
            Stype = "cone"
            x1, t2 = MCNPparams[0:2]
            t = math.sqrt(t2)
            dblsht = True

            if len(MCNPparams) == 3:
                dblsht = False
                sht = MCNPparams[2]
                if sht == 0.0:
                    dblsht = True

            if MCNPtype == "KX":
                p = FreeCAD.Vector(x1, 0.0, 0.0)
                v = X_vec
                if not dblsht:
                    if sht < 0:
                        v = negX_vec
            elif MCNPtype == "KY":
                p = FreeCAD.Vector(0.0, x1, 0.0)
                v = Y_vec
                if not dblsht:
                    if sht < 0:
                        v = negY_vec
            elif MCNPtype == "KZ":
                p = FreeCAD.Vector(0.0, 0.0, x1)
                v = Z_vec
                if not dblsht:
                    if sht < 0:
                        v = negZ_vec

            p = p.multiply(scale)
            params = (p, v, t, dblsht)

        elif MCNPtype in ["K/X", "K/Y", "K/Z"]:
            Stype = "cone"
            p = FreeCAD.Vector(MCNPparams[0:3])
            t2 = MCNPparams[3]
            t = math.sqrt(t2)
            dblsht = True

            if len(MCNPparams) == 5:
                dblsht = False
                sht = MCNPparams[4]
                if sht == 0.0:
                    dblsht = True

            if MCNPtype == "K/X":
                v = X_vec
                if not dblsht:
                    if sht < 0:
                        v = negX_vec
            elif MCNPtype == "K/Y":
                v = Y_vec
                if not dblsht:
                    if sht < 0:
                        v = negY_vec
            elif MCNPtype == "K/Z":
                v = Z_vec
                if not dblsht:
                    if sht < 0:
                        v = negZ_vec

            p = p.multiply(scale)
            params = (p, v, t, dblsht)

        elif MCNPtype in ["TX", "TY", "TZ"]:
            Stype = "torus"
            p = FreeCAD.Vector(MCNPparams[0:3])
            Ra, Rb, Rc = MCNPparams[3:6]

            if MCNPtype == "TX":
                v = X_vec
            elif MCNPtype == "TY":
                v = Y_vec
            elif MCNPtype == "TZ":
                v = Z_vec

            if scale != 1.0:
                Ra *= scale
                Rb *= scale
                Rc *= scale
                p = p.multiply(scale)

            params = (p, v, Ra, Rb, Rc)

        elif MCNPtype == "GQ" or MCNPtype == "SQ":
            Qparams = tuple(MCNPparams[0:10])

            if MCNPtype == "SQ":
                Stype, quadric = sq2params(Qparams)
            else:
                Stype, quadric = gq2params(Qparams)

            if Stype == "cylinder":
                # p = FreeCAD.Vector(quadric[0:3])
                # v = FreeCAD.Vector(quadric[3:6])
                # R = quadric[6]
                p, v, R = quadric
                if scale != 1.0:
                    R *= scale
                    p = p.multiply(scale)

                params = (p, v, R)

            elif Stype == "cylinder_elliptic":
                p, v, radii, raxes = quadric
                if scale != 1.0:
                    radii[0] *= scale
                    radii[1] *= scale
                    p = p.multiply(scale)
                params = (p, v, radii, raxes)

            elif Stype == "cylinder_hyperbolic":
                p, v, radii, raxes = quadric
                if scale != 1.0:
                    radii[0] *= scale
                    radii[1] *= scale
                    p = p.multiply(scale)
                params = (p, v, radii, raxes)

            elif Stype == "cone":
                # p = FreeCAD.Vector(quadric[0:3])
                # v = FreeCAD.Vector(quadric[3:6])
                # t = quadric[6]
                # dblsht = quadric[7]
                p, v, t, dblsht = quadric
                if scale != 1.0:
                    p = p.multiply(scale)
                params = (p, v, t, dblsht)

            elif Stype == "cone_elliptic":
                p, v, Ra, radii, raxes, dblsht = quadric
                if scale != 1.0:
                    Ra *= scale
                    radii[0] *= scale
                    radii[1] *= scale
                    p = p.multiply(scale)
                params = (p, v, Ra, radii, raxes, dblsht)

            elif Stype == "hyperboloid":
                p, v, radii, raxes, onesht = quadric
                if scale != 1.0:
                    radii[0] *= scale
                    radii[1] *= scale
                    p = p.multiply(scale)
                params = (p, v, radii, raxes, onesht)

            elif Stype == "ellipsoid":
                p, v, radii, raxes = quadric
                if scale != 1.0:
                    radii[0] *= scale
                    radii[1] *= scale
                    p = p.multiply(scale)
                params = (p, v, radii, raxes)

            elif Stype == "paraboloid":
                p, v, focal = quadric
                if scale != 1.0:
                    focal *= scale
                    p = p.multiply(scale)

                params = (p, v, focal)

            else:
                print(Stype)
                params = None
        #                get_quadric_surface(params)

        elif MCNPtype == "X":
            if len(MCNPparams) == 2:
                Stype = "plane"
                params = (X_vec, MCNPparams[0] * scale)
            elif len(MCNPparams) == 4:
                if (abs(MCNPparams[1] - MCNPparams[3])) > 1.0e-12:
                    Stype = "cone"
                    dblsht = False
                    t = (MCNPparams[3] - MCNPparams[1]) / (MCNPparams[2] - MCNPparams[0])
                    x = MCNPparams[0] - MCNPparams[1] / t
                    if (MCNPparams[0] - x) * (MCNPparams[2] - x) > 0:
                        p = FreeCAD.Vector(x, 0.0, 0.0)
                        if (MCNPparams[0] - x) > 0:
                            v = X_vec
                        else:
                            v = negX_vec
                        if scale != 1.0:
                            p *= scale
                        params = (p, v, abs(t), dblsht)
                elif abs(MCNPparams[1]) < 1.0e-12:
                    Stype = "plane"
                    if (abs(MCNPparams[0] - MCNPparams[2])) < 1.0e-12:
                        params = (X_vec, MCNPparams[0] * scale)
                else:
                    Stype = "cylinder"
                    if scale != 1.0:
                        p = p.multiply(scale)
                        R *= scale
                    params = (origin, X_vec, MCNPparams[1])
            else:
                print("not implemented surfaces defined by point with more than 2couples of value")

        elif MCNPtype == "Y":
            if len(MCNPparams) == 2:
                Stype = "plane"
                params = (Y_vec, MCNPparams[0] * scale)
            elif len(MCNPparams) == 4:
                if (abs(MCNPparams[1] - MCNPparams[3])) > 1.0e-12:
                    Stype = "cone"
                    dblsht = False
                    t = (MCNPparams[3] - MCNPparams[1]) / (MCNPparams[2] - MCNPparams[0])
                    y = MCNPparams[0] - MCNPparams[1] / t
                    if (MCNPparams[0] - y) * (MCNPparams[2] - y) > 0:
                        p = FreeCAD.Vector(0.0, y, 0.0)
                        if (MCNPparams[0] - y) > 0:
                            v = Y_vec
                        else:
                            v = negY_vec
                        if scale != 1.0:
                            p = p.multiply(scale)
                        params = (p, v, abs(t), dblsht)
                elif abs(MCNPparams[1]) < 1.0e-12:
                    Stype = "plane"
                    if (abs(MCNPparams[0] - MCNPparams[2])) < 1.0e-12:
                        params = (Y_vec, MCNPparams[0] * scale)
                else:
                    Stype = "cylinder"
                    if scale != 1.0:
                        p = p.multiply(scale)
                        R *= scale
                    params = (origin, Y_vec, MCNPparams[1])
            else:
                print("not implemented surfaces defined by point with more than 2couples of value")

        elif MCNPtype == "Z":
            if len(MCNPparams) == 2:
                Stype = "plane"
                params = (Z_vec, MCNPparams[0] * scale)
            elif len(MCNPparams) == 4:
                if (abs(MCNPparams[1] - MCNPparams[3])) > 1.0e-12:
                    Stype = "cone"
                    dblsht = False
                    t = (MCNPparams[3] - MCNPparams[1]) / (MCNPparams[2] - MCNPparams[0])
                    z = MCNPparams[0] - MCNPparams[1] / t
                    if (MCNPparams[0] - z) * (MCNPparams[2] - z) > 0:
                        p = FreeCAD.Vector(0.0, 0.0, z)
                        if (MCNPparams[0] - z) > 0:
                            v = Z_vec
                        else:
                            v = negZ_vec
                        if scale != 1.0:
                            p = p.multiply(scale)
                        params = (p, v, abs(t), dblsht)
                elif abs(MCNPparams[1]) < 1.0e-12:
                    Stype = "plane"
                    if (abs(MCNPparams[0] - MCNPparams[2])) < 1.0e-12:
                        params = (Z_vec, MCNPparams[0] * scale)
                else:
                    Stype = "cylinder"
                    if scale != 1.0:
                        p = p.multiply(scale)
                        R *= scale
                    params = (origin, Z_vec, MCNPparams[1])
            else:
                print("not implemented surfaces defined by point with more than 2couples of value")

        elif MCNPtype == "BOX":
            Stype = "box"
            p = FreeCAD.Vector(MCNPparams[0:3])
            v1 = FreeCAD.Vector(MCNPparams[3:6])
            v2 = FreeCAD.Vector(MCNPparams[6:9])
            v3 = FreeCAD.Vector(MCNPparams[9:12])
            if scale != 1.0:
                p = p.multiply(scale)
                v1 = v1.multiply(scale)
                v2 = v2.multiply(scale)
                v3 = v3.multiply(scale)
            params = (p, v1, v2, v3)

        elif MCNPtype == "RPP":
            Stype = "box"
            xmin, xmax, ymin, ymax, zmin, zmax = MCNPparams[0:6]
            lx = xmax - xmin
            ly = ymax - ymin
            lz = zmax - zmin
            p = FreeCAD.Vector(xmin, ymin, zmin)
            v1 = FreeCAD.Vector(lx, 0, 0)
            v2 = FreeCAD.Vector(0, ly, 0)
            v3 = FreeCAD.Vector(0, 0, lz)
            if scale != 1.0:
                p = p.multiply(scale)
                v1 = v1.multiply(scale)
                v2 = v2.multiply(scale)
                v3 = v3.multiply(scale)
            params = (p, v1, v2, v3)

        elif MCNPtype == "RCC":
            Stype = "can"
            p = FreeCAD.Vector(MCNPparams[0:3])
            v = FreeCAD.Vector(MCNPparams[3:6])
            R = MCNPparams[6]
            if scale != 1.0:
                p = p.multiply(scale)
                v = v.multiply(scale)
                R *= scale
            params = (p, v, R)

        elif MCNPtype == "REC":
            Stype = "ecan"
            p = FreeCAD.Vector(MCNPparams[0:3])
            v = FreeCAD.Vector(MCNPparams[3:6])
            majAxis = FreeCAD.Vector(MCNPparams[6:9])
            majRad = majAxis.Length
            majAxis.normalize()

            if len(MCNPparams) == 12:
                minAxis = FreeCAD.Vector(MCNPparams[9:12])
                minRad = minAxis.Length
                minAxis.normalize()
            else:
                minRad = MCNPparams[9]
                minAxis = v.cross(majAxis) / v.Length

            if scale != 1.0:
                p = p.multiply(scale)
                v = v.multiply(scale)
                majRad *= scale
                minRad *= scale

            if minRad > majRad:
                minRad, majRad = majRad, minRad
                minAxis, majAxis = majAxis, minAxis

            params = (p, v, [minRad, majRad], [minAxis, majAxis])

        elif MCNPtype == "TRC":
            Stype = "tcone"
            p = FreeCAD.Vector(MCNPparams[0:3])
            v = FreeCAD.Vector(MCNPparams[3:6])
            R1, R2 = MCNPparams[6:8]
            if scale != 1.0:
                p = p.multiply(scale)
                v = v.multiply(scale)
                R1 *= scale
                R2 *= scale
            params = (p, v, R1, R2)

        if Stype == "plane":
            surfaces[Sid] = Plane(number, params, trsf)
        elif Stype == "sphere":
            surfaces[Sid] = Sphere(number, params, trsf)
        elif Stype == "cylinder" or Stype == "can":
            surfaces[Sid] = Cylinder(number, params, trsf, Stype == "can")
        elif Stype == "cylinder_elliptic" or Stype == "ecan":
            surfaces[Sid] = EllipticCylinder(number, params, trsf, Stype == "ecan")
        elif Stype == "cylinder_hyperbolic":
            surfaces[Sid] = HyperbolicCylinder(number, params, trsf)
        elif Stype == "cone" or Stype == "tcone":
            surfaces[Sid] = Cone(number, params, trsf, Stype == "tcone")
        elif Stype == "cone_elliptic":
            surfaces[Sid] = EllipticCone(number, params, trsf)
        elif Stype == "hyperboloid":
            surfaces[Sid] = Hyperboloid(number, params, trsf)
        elif Stype == "ellipsoid":
            surfaces[Sid] = Ellipsoid(number, params, trsf)
        elif Stype == "paraboloid":
            surfaces[Sid] = Paraboloid(number, params, trsf)
        elif Stype == "torus":
            surfaces[Sid] = Torus(number, params, trsf)
        elif Stype == "box":
            surfaces[Sid] = Box(number, params, trsf)
        else:
            print("Undefined", Sid, Stype)
            print(MCNPtype, number, MCNPparams)

    return surfaces


def points_to_coeffs(scf):
    # mcnp implementation to convert 3 point plane to
    # plane parameters

    tpp = [0] * 4
    for i in range(1, 4):
        j = i % 3 + 1
        k = 6 - i - j
        k -= 1
        j -= 1
        tpp[i - 1] = (
            scf[j] * (scf[k + 3] - scf[k + 6]) + scf[j + 3] * (scf[k + 6] - scf[k]) + scf[j + 6] * (scf[k] - scf[k + 3])
        )
        tpp[3] += scf[i - 1] * (scf[j + 3] * scf[k + 6] - scf[j + 6] * scf[k + 3])

    xm = 0
    coeff = [0] * 4
    for i in range(1, 5):
        if xm == 0 and tpp[4 - i] != 0:
            xm = 1 / tpp[4 - i]
        coeff[4 - i] = tpp[4 - i] * xm

    # coeff [0:3] a,b,c plane parameters
    # coeff [3]   d plane parameter
    # normalization is d set to one if origin is not in the plane


def get_parabola_parameters(eVal, eVect, T, U):
    iaxis, comp = U[1]
    center = FreeCAD.Vector(T)
    axis = FreeCAD.Vector(eVect[iaxis][0])
    e1 = eVal[(iaxis + 1) % 3]
    focal = comp / (4 * e1)
    if focal < 0:
        focal = -focal
        axis = -axis
    return (center, axis, focal)


def get_cylinder_parameters(eVal, eVect, T, k, iaxis):

    other1 = (iaxis + 1) % 3
    other2 = (iaxis + 2) % 3

    eMin = eVal[other1]
    eMaj = eVal[other2]

    axis = FreeCAD.Vector(np.transpose(eVect)[iaxis])
    pos = FreeCAD.Vector(T)
    if abs(eMin - eMaj) < 1.0e-5:
        radius = float(np.sqrt(k / eMaj))
        return "cylinder", (pos, axis, radius)
    else:
        iMin = other1
        iMaj = other2
        if abs(eMin) < abs(eMaj):
            eMin, eMaj = eMaj, eMin
            iMin, iMaj = iMaj, iMin

        majorRad = float(np.sqrt(abs(k / eMaj)))
        minorRad = float(np.sqrt(abs(k / eMin)))
        minorAxis = FreeCAD.Vector(eVect.T[iMin])  # define axis in global geometry
        majorAxis = FreeCAD.Vector(eVect.T[iMaj])
        if np.sign(eMaj) == np.sign(eMin):
            return "cylinder_elliptic", (
                pos,
                axis,
                [minorRad, majorRad],
                [minorAxis, majorAxis],
            )
        else:
            if np.sign(k) == np.sign(eMaj):
                return "cylinder_hyperbolic", (
                    pos,
                    axis,
                    [minorRad, majorRad],
                    [minorAxis, majorAxis],
                )
            else:
                return "cylinder_hyperbolic", (
                    pos,
                    axis,
                    [majorRad, minorRad],
                    [majorAxis, minorAxis],
                )


def get_cone_parameters(eVal, eVect, T, iaxis):

    other1 = (iaxis + 1) % 3
    other2 = (iaxis + 2) % 3
    pos = FreeCAD.Vector(T)

    if abs(eVal[other1] - eVal[other2]) < 1e-5:
        axis = FreeCAD.Vector(np.transpose(eVect)[iaxis])
        tan = float(np.sqrt(-eVal[other1] / eVal[iaxis]))
        return "cone", (pos, axis, tan, True)
    else:
        for i in range(3):
            if np.sign(eVal[(i + 1) % 3]) == np.sign(eVal[(i + 2) % 3]):
                iaxis = i
                other1 = (iaxis + 1) % 3
                other2 = (iaxis + 2) % 3
                break

        axis = FreeCAD.Vector(np.transpose(eVect)[iaxis])
        minAxis = FreeCAD.Vector(np.transpose(eVect)[other1])
        majAxis = FreeCAD.Vector(np.transpose(eVect)[other2])
        Ra = abs(1 / eVal[iaxis])
        Rmin = abs(1 / eVal[other1])
        Rmaj = abs(1 / eVal[other2])

        if Rmin > Rmaj:
            Rmin, Rmaj = Rmaj, Rmin
            minAxis, majAxis = majAxis, minAxis

        return "cone_elliptic", (pos, axis, Ra, [Rmin, Rmaj], [minAxis, majAxis], True)


def get_hyperboloid_parameters(eVal, eVect, T, k, iaxis):
    cylTan = 1e3
    coneRad = 0.1

    ellipsoid = False
    if iaxis is None:
        iaxis = np.argmin(np.abs(eVal))
        ellipsoid = True

    other1 = (iaxis + 1) % 3
    other2 = (iaxis + 2) % 3
    Rad1 = float(np.sqrt(abs(k / eVal[other1])))
    Rad2 = float(np.sqrt(abs(k / eVal[other2])))
    other = other1 if Rad1 > Rad2 else other2

    majorRad = float(np.sqrt(abs(k / eVal[iaxis])))
    minorRad = float(np.sqrt(abs(k / eVal[other])))
    oneSheet = np.sign(k) != np.sign(eVal[iaxis])

    axis = FreeCAD.Vector(np.transpose(eVect)[iaxis])
    pos = FreeCAD.Vector(T)
    minorAxis = FreeCAD.Vector(eVect.T[other])  # define axis in global geometry
    majorAxis = FreeCAD.Vector(eVect.T[iaxis])

    t = majorRad / minorRad

    if t > cylTan and oneSheet:
        return get_cylinder_parameters(eVal, eVect, T, k, iaxis)
    elif minorRad < coneRad:
        return get_cone_parameters(eVal, eVect, T, iaxis)
    else:
        if ellipsoid:
            print("ellipical hyperboloid not implemented")
            print(f"single radius from {minorRad} eigen Value will be used")
        return "hyperboloid", (
            pos,
            axis,
            [minorRad, majorRad],
            [minorAxis, majorAxis],
            oneSheet,
        )


def get_ellipsoid_parameters(eVal, eVect, T, k):

    cylTan = 1e3
    iaxis = None
    for i in range(3):
        if abs(eVal[i] - eVal[(i + 1) % 3]) < 1e-3:
            iaxis = (i + 2) % 3
            break

    if iaxis is None:
        print(eVal)
        print("cannot produce three radii Ellipsoid")
        return "ellipsoid_general", None

    other1 = (iaxis + 1) % 3
    iMaj = iaxis
    iMin = other1

    eMaj = eVal[iaxis]
    eMin = eVal[other1]

    if eMin < eMaj:
        eMin, eMaj = eMaj, eMin
        iMin, iMaj = iMaj, iMin

    RMaj = float(np.sqrt(abs(k / eMaj)))
    RMin = float(np.sqrt(abs(k / eMin)))
    majorAxis = FreeCAD.Vector(np.transpose(eVect)[iMaj])
    minorAxis = FreeCAD.Vector(np.transpose(eVect)[iMin])
    pos = FreeCAD.Vector(T)

    t = RMaj / RMin
    if t > cylTan:
        return get_cylinder_parameters(eVal, eVect, T, k, iMaj)
    else:
        return "ellipsoid", (pos, iaxis, [RMin, RMaj], [minorAxis, majorAxis])


def getGQAxis(eVal, k):

    # check if there is two equal eigenValues
    iaxis = None
    for i in range(3):
        if abs(eVal[i] - eVal[(i + 1) % 3]) < 1e-5:
            iaxis = (i + 2) % 3
            break

    if iaxis is None:
        iaxis = np.argmin(np.abs(eVal))

    e0 = eVal[iaxis]
    e1 = eVal[(iaxis + 1) % 3]
    e2 = eVal[(iaxis + 2) % 3]

    if k == 0:  # k == 0
        if e0 == 0:  # e1*X^2 + e2*Y^2             = 0    Intersecting  planes (real or imaginary)
            ek = None
        elif np.sign(e0) == np.sign(e1) and np.sign(e1) == np.sign(
            e2
        ):  # e1*X^2 + e2*Y^2 + e0*Z^2    = 0    Imaginary ellipsoid
            ek = None
        else:  # e1*X^2 + e2*Y^2 - e0*Z^2    = 0    Elliptic Cone
            ek = (-1, 0)

    elif np.sign(k) == np.sign(e1):  # e1 and k same sign  (e1 > 0)
        if e0 == 0:
            if np.sign(e1) == np.sign(e2):  # e1*X^2 + e2*Y^2          + |k| = 0  Imaginary Elliptic cylinder
                ek = None
            else:  # e1*X^2 - e2*Y^2          + |k| = 0  Hyperpolic cylinder
                ek = (0, -1)
        elif np.sign(e0) == np.sign(e1) and np.sign(e1) == np.sign(
            e2
        ):  # e1*X^2 + e2*Y^2 + e0*Z^2 + |k| = 0  Imaginary ellipsoid
            ek = None
        else:  # e1*X^2 + e2*Y^2 - e0*Z^2 + |k| = 0  Hyperboloid
            ek = (-1, 1)

    else:  # e1 and k different sign
        if e0 == 0:  # e1*X^2 + e2*Y^2          - |k| = 0  Elliptic cylinder
            ek = (0, -1)
        elif np.sign(e0) == np.sign(e1) and np.sign(e1) == np.sign(e2):  # e1*X^2 + e2*Y^2 + e0*Z^2 - |k| = 0  Ellipsoid
            ek = (1, -1)
        else:  # e1*X^2 + e2*Y^2 - e0*Z^2 - |k| = 0  Hyperbpoloid
            ek = (-1, 1)

    return iaxis, ek


def sq2params(params):

    a, b, c, d, e, f, g = params[0:7]
    x, y, z = params[7:10]

    gqa, gqb, gqc = a, b, c
    gqd, gqe, gqf = 0, 0, 0
    gqg = 2 * (d - a * x)
    gqh = 2 * (e - b * y)
    gqj = 2 * (f - c * z)
    gqk = g + a * x * x + b * y * y + c * z * z - 2 * d * x - 2 * e * y - 2 * f * z

    return gq2params((gqa, gqb, gqc, gqd, gqe, gqf, gqg, gqh, gqj, gqk))


def gq2params(x):
    # intial matrix: A
    #      a f g u
    #      f b h v
    #      g h c w
    #      u v w d

    # matrix displacement : D
    #      1 0 0 alpha
    #      0 1 0 beta
    #      0 0 1 gamma
    #      0 0 0 1

    #   Transpose(D) * A * D = mat3
    #   matrix mat3
    #      a f g 0
    #      f b h 0
    #      g h c 0
    #      0 0 0 (d + u*alpha + v*beta + w*gamma)

    #  with vector(alpha,beta,gamma) solution of
    #      a f h     alpha      -u
    #      f b g  *  beta   =   -v
    #      h g c     gamma      -w

    mat3 = np.array(
        [
            [x[0], x[3] / 2, x[5] / 2],
            [x[3] / 2, x[1], x[4] / 2],
            [x[5] / 2, x[4] / 2, x[2]],
        ]
    )
    X = np.array((x[6] / 2, x[7] / 2, x[8] / 2))

    # mat4 = np.array( [[x[0],x[3]/2,x[5]/2,x[6]/2], \
    #                  [x[3]/2,x[1],x[4]/2,x[7]/2], \
    #                  [x[5]/2,x[4]/2,x[2],x[8]/2], \
    #                  [x[6]/2,x[7]/2,x[8]/2,x[9]]] )

    # eigenValues and Vector
    eVal, vect = LA.eigh(mat3)
    XD = np.matmul(X, vect)  # X in diagonalised base

    Dinv = np.where(abs(eVal) < 1e-8, eVal, 1 / eVal)  # get inverse eigen value where eigen< 1e-8
    zero = (abs(eVal) < 1e-8).nonzero()  # index in eigen value vector where eigen < 1e-8
    TD = -XD * Dinv  # Translation vector in diagonalized base

    k = np.matmul(TD, XD) + x[9]

    if len(zero) != 0:
        iz = zero[0]
        comp = 2 * XD[iz]
        if (
            abs(comp) > 1e-6
        ):  # zero eigenvalue but corresponding component in XD vector is non zero => paraboloid Curve => the k/comp value is the translation in this component direction
            TD[iz] = -k / comp
            U = (k, (iz, comp))
        else:
            U = (k, None)
    else:
        U = (k, None)

    T = np.matmul(TD, vect.T)

    return conicSurface(eVal, vect, T, U)


def conicSurface(eVal, vect, T, U):

    # paraboloid
    if U[1] is not None:
        params = get_parabola_parameters(eVal, vect, T, U)
        stype = "paraboloid"
        return stype, params

    # other conics
    k = U[0]
    iaxis, ek = getGQAxis(eVal, k)
    if ek == (0, -1):
        # Cylinder
        stype, params = get_cylinder_parameters(eVal, vect, T, -k, iaxis)
    elif ek == (-1, 0):
        # cone
        stype, params = get_cone_parameters(eVal, vect, T, iaxis)
    elif ek == (-1, 1):
        # hyperboloid
        # function can return cylinder or cone if hyperboiloid can be aproximated by such surfaces (within some criterion)
        stype, params = get_hyperboloid_parameters(eVal, vect, T, -k, iaxis)
    elif ek == (1, -1):
        # ellipsoid
        # function can return cylinder or cone if hyperboiloid can be aproximated by such surfaces (within some criterion)
        stype, params = get_ellipsoid_parameters(eVal, vect, T, -k)
    else:
        stype = "unknown"
        params = None
        print("No 2nd order or real surfaces")
    return stype, params



import logging

import FreeCAD
import Part
from tqdm import tqdm

from ..loadfile import load_functions as LF
from ..utils.basic_functions_part1 import is_opposite
from ..utils.boolean_function import BoolSequence
from ..utils.functions import GeounedSolid, GeounedSurface
from . import void_functions as VF
from .void_box_class import VoidBox

logger = logging.getLogger("general_logger")


def void_generation(
    MetaList,
    EnclosureList,
    Surfaces,
    UniverseBox,
    setting,
    init,
    options,
    tolerances,
    numeric_format,
):
    voidList = []

    if EnclosureList:
        NestedEnclosure = LF.set_enclosure_levels(EnclosureList)
        VF.assignEnclosure(MetaList, NestedEnclosure)

        # add to Metalist Level 1 enclosures, remove from list material cells totally embedded in Level 1 enclosures
        newMetaList = VF.select_solids(MetaList, NestedEnclosure[0], UniverseBox)
    else:
        newMetaList = MetaList[:]
        NestedEnclosure = []

    Box = Part.makeBox(
        UniverseBox.XLength,
        UniverseBox.YLength,
        UniverseBox.ZLength,
        FreeCAD.Vector(UniverseBox.XMin, UniverseBox.YMin, UniverseBox.ZMin),
        FreeCAD.Vector(0, 0, 1),
    )

    EnclosureBox = GeounedSolid(None, Box)
    if setting.voidMat:
        voidMat = setting.voidMat
        EnclosureBox.set_material(voidMat[0], voidMat[1], voidMat[2])

    # get voids in 0 Level Enclosure (original Universe)
    # if exist Level 1 enclosures are considered as material cells
    logger.info("Build Void highest enclosure")

    voids = get_void_def(
        newMetaList,
        Surfaces,
        EnclosureBox,
        setting,
        options,
        tolerances,
        numeric_format,
        Lev0=True,
    )
    voidList.append(voids)

    # Perform enclosure void
    # Loop until the lowest enclosure level

    for i, Level in enumerate(NestedEnclosure):

        logger.info("Build Void highest enclosure")
        for j, encl in enumerate(Level):
            if encl.CellType == "envelope":
                continue
            newMetaList = VF.select_solids(MetaList, encl.SonEnclosures, encl)
            logger.info(f"Build Void enclosure {j} in enclosure level {i + 1}")
            # select solids overlapping current enclosure "encl", and lower level enclosures
            voids = get_void_def(
                newMetaList,
                Surfaces,
                encl,
                setting,
                options,
                tolerances,
                numeric_format,
            )
            voidList.append(voids)

    voidList.append(set_graveyard_cell(Surfaces, UniverseBox, options, tolerances, numeric_format))

    return VF.update_void_list(init, voidList, NestedEnclosure, setting.sort_enclosure)


def get_void_def(
    MetaList,
    Surfaces,
    Enclosure,
    setting,
    options,
    tolerances,
    numeric_format,
    Lev0=False,
):

    maxsurf = setting.maxSurf
    maxbracket = setting.maxBracket
    minSize = setting.minVoidSize

    if "full" in setting.simplify.lower():
        simplifyVoid = "full"
    elif "void" in setting.simplify.lower():
        simplifyVoid = "diag"
    else:
        simplifyVoid = "no"

    if Lev0:
        Universe = VoidBox(MetaList, Enclosure.BoundBox)
    else:
        Universe = VoidBox(MetaList, Enclosure.CADSolid)

    Initial = [Universe]
    VoidDef = []
    iloop = 0
    while iloop < 50:
        Temp = []
        iloop += 1
        nvoid = len(Initial)
        logger.info("Loop, Box to Split :{iloop}, {nvoid}")

        for iz, z in enumerate(tqdm(Initial, desc=f"Void Generation Loop: {iloop}")):
            nsurfaces, nbrackets = z.get_numbers()
            logger.info(f"{iloop} {iz + 1}/{nvoid} {nsurfaces} {nbrackets}")

            if nsurfaces > maxsurf and nbrackets > maxbracket:
                newspace = z.split(minSize)
            else:
                newspace = None

            if type(newspace) is tuple:
                Temp.extend(newspace)
            else:
                #           if len(z.Objects) >= 50 : z.refine()
                boxDim = (
                    z.BoundBox.XMin * 0.1,
                    z.BoundBox.XMax * 0.1,
                    z.BoundBox.YMin * 0.1,
                    z.BoundBox.YMax * 0.1,
                    z.BoundBox.ZMin * 0.1,
                    z.BoundBox.ZMax * 0.1,
                )

                logger.info(f"build complementary {iloop} {iz}")

                cell, CellIn = z.get_void_complementary(Surfaces, options, tolerances, numeric_format, simplify=simplifyVoid)
                if cell is not None:
                    VoidCell = (cell, (boxDim, CellIn))
                    VoidDef.append(VoidCell)

        Initial = Temp
        if len(Temp) == 0:
            break

    voidList = []
    for i, vcell in enumerate(VoidDef):
        mVoid = GeounedSolid(i)
        mVoid.Void = True
        mVoid.CellType = "void"
        mVoid.set_definition(vcell[0], simplify=True)
        mVoid.set_material(Enclosure.Material, Enclosure.Rho, Enclosure.MatInfo)
        mVoid.set_dilution(Enclosure.Dilution)

        mVoid.__commentInfo__ = vcell[1]

        voidList.append(mVoid)

    return voidList


def set_graveyard_cell(Surfaces, UniverseBox, options, tolerances, numeric_format):
    Universe = VoidBox([], UniverseBox)

    externalBox = get_universe_complementary(Universe, Surfaces, options, tolerances, numeric_format)
    center = UniverseBox.Center
    radius = 0.51 * UniverseBox.DiagonalLength
    sphere = GeounedSurface(("Sphere", (center, radius)), UniverseBox)
    id, exist = Surfaces.add_sphere(sphere, tolerances)

    sphdef = BoolSequence(str(-id))
    sphdef.operator = "AND"
    sphdef.append(externalBox)

    notsph = BoolSequence(str(id))

    mVoidSphIn = GeounedSolid(0)
    mVoidSphIn.Void = True
    mVoidSphIn.CellType = "void"
    mVoidSphIn.set_definition(sphdef)
    mVoidSphIn.set_material(0, 0, "Graveyard_in")
    mVoidSphIn.__commentInfo__ = None

    mVoidSphOut = GeounedSolid(1)
    mVoidSphOut.Void = True
    mVoidSphOut.CellType = "void"
    mVoidSphOut.set_definition(notsph)
    mVoidSphOut.set_material(0, 0, "Graveyard")
    mVoidSphOut.__commentInfo__ = None

    return (mVoidSphIn, mVoidSphOut)


# TODO check this is being used
def get_universe_complementary(Universe, Surfaces, options, tolerances, numeric_format):
    Def = BoolSequence(operator="OR")
    for p in Universe.get_bound_planes():
        id, exist = Surfaces.add_plane(p, options, tolerances, numeric_format, False)
        if not exist:
            Def.elements.append(-id)
        else:
            s = Surfaces.get_surface(id)
            if is_opposite(p.Surf.Axis, s.Surf.Axis):
                Def.elements.append(id)
            else:
                Def.elements.append(-id)
    return Def


def void_comment_line(CellInfo):
    boxDef, cellIn = CellInfo
    cells = ", ".join(map(str, cellIn))
    box = ", ".join(f"{num:.3f}" for num in boxDef)
    line = f"Automatic Generated Void Cell. Enclosure({box})\n"
    line += f"Enclosed cells : ({cells})\n"
    return line



############################
# Module for Cell definiton #
#############################
import logging
import math

import FreeCAD
import Part

from ..utils import basic_functions_part2 as BF
from ..utils import functions as UF
from ..utils import geometry_gu as GU
from ..utils.basic_functions_part1 import (
    is_in_line,
    is_opposite,
    is_parallel,
    is_same_value,
    sign_plane,
)
from ..utils.boolean_function import BoolSequence, insert_in_sequence
from ..utils.boolean_solids import build_c_table_from_solids, remove_extra_surfaces
from ..utils.functions import GeounedSurface

logger = logging.getLogger("general_logger")


# TODO rename this function as there are two with the same name
def get_id(facein, surfaces, options, tolerances, numeric_format):

    surfin = str(facein)
    if surfin == "<Plane object>":
        if is_parallel(facein.Axis, FreeCAD.Vector(1, 0, 0), tolerances.pln_angle):
            p = "PX"
        elif is_parallel(facein.Axis, FreeCAD.Vector(0, 1, 0), tolerances.pln_angle):
            p = "PY"
        elif is_parallel(facein.Axis, FreeCAD.Vector(0, 0, 1), tolerances.pln_angle):
            p = "PZ"
        else:
            p = "P"

        for s in surfaces[p]:
            if BF.is_same_plane(
                facein,
                s.Surf,
                options=options,
                tolerances=tolerances,
                numeric_format=numeric_format,
            ):
                return s.Index

    elif surfin == "<Cylinder object>":
        for s in surfaces["Cyl"]:
            if BF.is_same_cylinder(
                facein,
                s.Surf,
                options=options,
                tolerances=tolerances,
                numeric_format=numeric_format,
            ):
                return s.Index

    elif surfin == "<Cone object>":
        for s in surfaces["Cone"]:
            if BF.is_same_cone(
                facein,
                s.Surf,
                dtol=tolerances.kne_distance,
                atol=tolerances.kne_angle,
                rel_tol=tolerances.relativeTol,
            ):
                return s.Index

    elif surfin[0:6] == "Sphere":
        for s in surfaces["Sph"]:
            if BF.is_same_sphere(facein, s.Surf, tolerances.sph_distance, rel_tol=tolerances.relativeTol):
                return s.Index

    elif surfin == "<Toroid object>":
        for s in surfaces["Tor"]:
            if BF.is_same_torus(
                facein,
                s.Surf,
                dtol=tolerances.tor_distance,
                atol=tolerances.tor_angle,
                rel_tol=tolerances.relativeTol,
            ):
                return s.Index

    return 0


def is_inverted(solid):

    face = solid.Faces[0]

    # u=(face.Surface.bounds()[0]+face.Surface.bounds()[1])/2.0 # entre 0 y 2pi si es completo
    # v=face.Surface.bounds()[0]+(face.Surface.bounds()[3]-face.Surface.bounds()[2])/3.0 # a lo largo del eje
    parameter_range = face.ParameterRange
    u = (parameter_range[1] + parameter_range[0]) / 2.0
    v = (parameter_range[3] + parameter_range[2]) / 2.0

    if str(face.Surface) == "<Cylinder object>":
        dist1 = face.Surface.value(u, v).distanceToLine(face.Surface.Center, face.Surface.Axis)
        dist2 = (
            face.Surface.value(u, v)
            .add(face.Surface.normal(u, v).multiply(1.0e-6))
            .distanceToLine(face.Surface.Center, face.Surface.Axis)
        )
        if (dist2 - dist1) < 0.0:
            # The normal of the cylinder is going inside
            return True
    elif str(face.Surface) == "<Cone object>":
        dist1 = face.Surface.value(u, v).distanceToLine(face.Surface.Apex, face.Surface.Axis)
        dist2 = (
            face.Surface.value(u, v)
            .add(face.Surface.normal(u, v).multiply(1.0e-6))
            .distanceToLine(face.Surface.Apex, face.Surface.Axis)
        )
        if (dist2 - dist1) < 0.0:
            # The normal of the cylinder is going inside
            return True
    # MIO
    elif str(face.Surface)[0:6] == "Sphere":
        # radii = point - center
        radii = face.Surface.value(u, v).add(face.Surface.Center.multiply(-1))
        radii_b = face.Surface.value(u, v).add(face.Surface.normal(u, v).multiply(1.0e-6)).add(face.Surface.Center.multiply(-1))
        # radii_b  = radii.add( face.Surface.normal(u,v).multiply(1.0e-6) )
        if (radii_b.Length - radii.Length) < 0.0:
            # An increasing of the radii vector in the normal direction decreases the radii: oposite normal direction
            return True

    elif str(face.Surface) == "<Plane object>":
        dist1 = face.CenterOfMass.distanceToPoint(solid.BoundBox.Center)
        dist2 = face.CenterOfMass.add(face.normalAt(u, v).multiply(1.0e-6)).distanceToPoint(solid.BoundBox.Center)
        point2 = face.CenterOfMass.add(face.normalAt(u, v).multiply(1.0e-6))
        if solid.isInside(point2, 1e-7, False):
            return True

    return False


def gen_plane(face, solid, tolerances):
    """Generate an additional plane when convex surfaces of second order are presented as a face of the solid"""

    surf = face.Surface
    if str(surf) == "<Cylinder object>":
        return gen_plane_cylinder(face, solid, tolerances)
    if str(surf) == "<Cone object>":
        return gen_plane_cone(face, solid, tolerances)
    if str(surf) == "Sphere":
        return gen_plane_sphere(face, solid)


def get_closed_ranges(solid, face_index):

    u_nodes = []
    for index in face_index:
        URange = solid.Faces[index].ParameterRange
        u_nodes.append((URange[0], index))
        u_nodes.append((URange[1], index))
    u_nodes.sort()

    closed_range = get_intervals(u_nodes)

    a_min = closed_range[0][0][0]
    a_max = closed_range[-1][1][0]

    if abs(a_max - a_min - 2.0 * math.pi) < 1e-2:
        if len(closed_range) == 1:
            closed_face = True
        else:
            endPoint = (closed_range[-1][0][0] - 2 * math.pi, closed_range[-1][0][1])
            closed_range[0][0] = endPoint
            closed_range[0][2].update(closed_range[-1][2])
            del closed_range[-1]

            if len(closed_range) == 1:
                if abs(closed_range[0][1][0] - closed_range[0][0][0] - 2.0 * math.pi) < 1e-2:
                    closed_face = True
                else:
                    closed_face = False
            else:
                closed_face = False
    else:
        closed_face = False
    return closed_range, closed_face


def get_intervals(u_nodes):
    closed_ranges = []
    pos_min = dict()
    pos_max = dict()
    for i, node in enumerate(u_nodes):
        if node[1] not in pos_min.keys():
            pos_min[node[1]] = i
        else:
            pos_max[node[1]] = i

    u_min = u_nodes[0]
    i_pos = pos_max[u_min[1]]

    while True:
        x = u_nodes[i_pos]
        end = True
        for i in range(i_pos + 1, len(u_nodes)):
            mxt_int = u_nodes[i][1]
            if (
                u_nodes[pos_min[mxt_int]][0] - x[0]
            ) < 1e-5:  # x pos is > min boundary of the next inteval inside precision 1e-5
                i_pos = pos_max[mxt_int]
                end = False
                break

        if end:
            u_max = x
            closed_ranges.append([u_min, u_max])
            i_pos += 1
            if i_pos < len(u_nodes):
                u_min = u_nodes[i_pos]
                i_pos = pos_max[u_min[1]]
            else:
                break

    for closed_range in closed_ranges:
        index = set()
        xmin = closed_range[0][0]
        xmax = closed_range[1][0]
        for interval in u_nodes:
            x = interval[0]
            if (xmin - x) < 1.0e-5 and (x - xmax) < 1.0e-5:
                index.add(interval[1])
        closed_range.append(index)

    return closed_ranges


def get_u_value_boundary(solid, face_index, my_index):

    face_u_ranges, closed_face = get_closed_ranges(solid, face_index)
    if closed_face:
        return None, None

    for face_u_range in face_u_ranges:
        if my_index in face_u_range[2]:
            u_min, u_max = face_u_range[0:2]
            return u_min, u_max


def gen_plane_sphere(face, solid):
    same_faces = []
    same_faces.append(face)

    for f in solid.Faces:
        if f.isEqual(face) or str(f.Surface) != "Sphere":
            continue
        if f.Surface.Center == face.Surface.Center and f.Surface.Radius == face.Surface.Radius:
            # print 'Warning: coincident sphere faces are the same'
            for f2 in same_faces:
                if f.__face__.distToShape(f2.__face__)[0] < 1e-6:
                    same_faces.append(f)
                    break

    # print same_faces
    normal = FreeCAD.Vector(0, 0, 0)
    for f in same_faces:
        normal += f.Area * (f.CenterOfMass - face.Surface.Center)
    normal.normalize()
    tmp_plane = Part.Plane(face.Surface.Center, normal).toShape()

    dmin = 2 * face.Surface.Radius
    for f in same_faces:
        dist = tmp_plane.distToShape(f.__face__)[0]
        dmin = min(dmin, dist)

    if dmin > 1e-6:
        center = face.Surface.Center + 0.95 * dmin * normal
        plane = Part.Plane(center, normal).toShape()
    else:
        plane = None
    return plane


def gen_plane_cylinder(face, solid, tolerances):

    surf = face.Surface
    rad = surf.Radius

    if str(surf) != "<Cylinder object>":
        return None

    my_index = solid.Faces.index(face)
    face_index = [my_index]

    for i, face2 in enumerate(solid.Faces):
        if face2.Area < tolerances.min_area:
            logger.warning(
                f"surface {str(surf)} removed from cell definition. Face area < Min area ({face2.Area} < {tolerances.min_area})"
            )
            continue
        if str(face2.Surface) == "<Cylinder object>" and not (face2.isEqual(face)):
            if (
                face2.Surface.Axis.isEqual(face.Surface.Axis, 1e-5)
                and face2.Surface.Radius == rad
                and is_in_line(face2.Surface.Center, face.Surface.Axis, face.Surface.Center)
            ):
                # print 'Warning: coincident cylinder faces are the same'
                face_index.append(i)

    u_min, u_max = get_u_value_boundary(solid, face_index, my_index)
    if u_min is None:
        return None

    u_1, i1 = u_min
    u_2, i2 = u_max

    v_1 = solid.Faces[i1].ParameterRange[2]
    v_2 = solid.Faces[i2].ParameterRange[2]

    p1 = solid.Faces[i1].valueAt(u_1, v_1)
    p2 = solid.Faces[i2].valueAt(u_2, v_2)

    if p1.isEqual(p2, 1e-5):
        logger.error("Error in the additional place definition")
        return None

    normal = p2.sub(p1).cross(face.Surface.Axis)
    plane = Part.Plane(p1, normal).toShape()

    return plane


def gen_plane_cone(face, solid, tolerances):

    Surf = face.Surface
    if str(Surf) != "<Cone object>":
        return None

    myIndex = solid.Faces.index(face)
    face_index = [myIndex]

    for i, face2 in enumerate(solid.Faces):
        if face2.Area < tolerances.min_area:
            logger.warning(
                f"{str(Surf)} surface removed from cell definition. Face area < Min area ({face2.Area} < {tolerances.min_area})"
            )
            continue
        if str(face2.Surface) == "<Cone object>" and not (face2.isEqual(face)):
            if (
                face2.Surface.Axis.isEqual(face.Surface.Axis, 1e-5)
                and face2.Surface.Apex.isEqual(face.Surface.Apex, 1e-5)
                and (face2.Surface.SemiAngle - face.Surface.SemiAngle) < 1e-6
            ):
                face_index.append(i)

    u_min, u_max = get_u_value_boundary(solid, face_index, myIndex)
    if u_min is None:
        return None

    u_1, i1 = u_min
    u_2, i2 = u_max

    v_1 = solid.Faces[i1].ParameterRange[2]
    v_2 = solid.Faces[i2].ParameterRange[2]

    p1 = solid.Faces[i1].valueAt(u_1, v_1)
    p2 = solid.Faces[i2].valueAt(u_2, v_2)

    if p1.isEqual(p2, 1e-5):
        logger.error("in the additional place definition")
        return None

    plane = Part.Plane(p1, p2, face.Surface.Apex).toShape()

    return plane


def gen_torus_annex_u_planes(face, u_params, tolerances):

    if is_parallel(face.Surface.Axis, FreeCAD.Vector(1, 0, 0), tolerances.tor_angle):
        axis = FreeCAD.Vector(1, 0, 0)
    elif is_parallel(face.Surface.Axis, FreeCAD.Vector(0, 1, 0), tolerances.tor_angle):
        axis = FreeCAD.Vector(0, 1, 0)
    elif is_parallel(face.Surface.Axis, FreeCAD.Vector(0, 0, 1), tolerances.tor_angle):
        axis = FreeCAD.Vector(0, 0, 1)

    center = face.Surface.Center
    p1 = face.valueAt(u_params[0], 0.0)
    p2 = face.valueAt(u_params[1], 0.0)
    pmid = face.valueAt(0.5 * (u_params[0] + u_params[1]), 0.0)

    if is_same_value(abs(u_params[1] - u_params[0]), math.pi, tolerances.value):
        d = axis.cross(p2 - p1)
        d.normalize()
        if d.dot(pmid - center) < 0:
            d = -d
        return (
            (center, d, face.Surface.MajorRadius, face.Surface.MajorRadius),
            None,
        ), False

    elif u_params[1] - u_params[0] < math.pi:
        d = axis.cross(p2 - p1)
        d.normalize()
        if d.dot(pmid - center) < 0:
            d = -d
        return (
            (center, d, face.Surface.MajorRadius, face.Surface.MajorRadius),
            None,
        ), False

    else:
        d1 = axis.cross(p1)
        d1.normalize()
        if d1.dot(pmid - center) < 0:
            d1 = -d1

        d2 = axis.cross(p2)
        d2.normalize()
        if d2.dot(pmid - center) < 0:
            d2 = -d2

        return (
            (center, d1, face.Surface.MajorRadius, face.Surface.MajorRadius),
            (center, d2, face.Surface.MajorRadius, face.Surface.MajorRadius),
        ), True  # (d1 : d2)


def gen_torus_annex_v_surface(face, v_params, tolerances, force_cylinder=False):
    if is_parallel(face.Surface.Axis, FreeCAD.Vector(1, 0, 0), tolerances.tor_angle):
        axis = FreeCAD.Vector(1, 0, 0)
    elif is_parallel(face.Surface.Axis, FreeCAD.Vector(0, 1, 0), tolerances.tor_angle):
        axis = FreeCAD.Vector(0, 1, 0)
    elif is_parallel(face.Surface.Axis, FreeCAD.Vector(0, 0, 1), tolerances.tor_angle):
        axis = FreeCAD.Vector(0, 0, 1)

    p1 = face.valueAt(0.0, v_params[0]) - face.Surface.Center
    z1 = p1.dot(axis)
    d1 = p1.cross(axis).Length

    p2 = face.valueAt(0.0, v_params[1]) - face.Surface.Center
    z2 = p2.dot(axis)
    d2 = p2.cross(axis).Length

    if is_same_value(z1, z2, tolerances.distance):
        surf_type = "Plane"
        center = face.Surface.Center + z1 * axis
        v_mid = (v_params[0] + v_params[1]) * 0.5
        p_mid = face.valueAt(0, v_mid) - face.Surface.Center
        if p_mid.dot(axis) < z1:
            in_surf = True
        else:
            in_surf = False
        return (
            (center, axis, face.Surface.MajorRadius, face.Surface.MajorRadius),
            surf_type,
            in_surf,
        )

    elif is_same_value(d1, d2, tolerances.distance) or force_cylinder:
        surf_type = "Cylinder"
        radius = min(d1, d2)
        center = face.Surface.Center
        if is_same_value(d1, face.Surface.MajorRadius, tolerances.distance):
            v_mid = (v_params[0] + v_params[1]) * 0.5
            p_mid = face.valueAt(0, v_mid) - center
            if p_mid.cross(axis).Length < face.Surface.MajorRadius:
                in_surf = True
            v_mid = (v_params[0] + v_params[1]) * 0.5
            p_mid = face.valueAt(0, v_mid) - center
            if p_mid.cross(axis).Length < face.Surface.MajorRadius:
                in_surf = True
                radius = max(d1, d2)
            else:
                in_surf = False
        else:
            if d1 < face.Surface.MajorRadius:
                in_surf = True
                radius = max(d1, d2)
            else:
                in_surf = False
        return (center, axis, radius, face.Surface.MinorRadius), surf_type, in_surf

    else:
        surf_type = "Cone"
        za = (z2 * d1 - z1 * d2) / (d1 - d2)
        apex = face.Surface.Center + za * axis
        semi_angle = abs(math.atan(d1 / (z1 - za)))

        cone_axis = axis if (z1 - za) > 0.0 else -axis

        v_mid = (v_params[0] + v_params[1]) * 0.5
        p_mid = face.valueAt(0, v_mid) - face.Surface.Center
        z_mid = p_mid.dot(axis)
        d_mid = p_mid.cross(axis).Length

        d_cone = d1 * (z_mid - za) / (z1 - za)
        in_surf = True if d_mid < d_cone else False

        return (
            (
                apex,
                cone_axis,
                semi_angle,
                face.Surface.MinorRadius,
                face.Surface.MajorRadius,
            ),
            surf_type,
            in_surf,
        )


def cellDef(meta_obj, surfaces, universe_box, options, tolerances, numeric_format):

    solids = meta_obj.Solids
    del_list = []

    piece_def = BoolSequence(operator="OR")
    iece_obj = []
    cones = set()
    for isol, solid in enumerate(solids):
        surf_piece = []
        surf_obj = []
        extra_plane_reverse = dict()

        flag_inv = is_inverted(solid)
        solid_gu = GU.SolidGu(solid, tolerances=tolerances)
        last_torus = -1
        for iface, face in enumerate(solid_gu.Faces):
            surface_type = str(face.Surface)
            if abs(face.Area) < tolerances.min_area:
                logger.warning(
                    f"{surface_type} surface removed from cell definition. Face area < Min area ({face.Area} < {tolerances.min_area})"
                )
                continue
            if face.Area < 0:
                logger.warning("Negative surface Area")
            if face.Orientation not in ("Forward", "Reversed"):
                continue
            if flag_inv:
                orient_temp = face.Orientation
                if orient_temp == "Forward":
                    orient = "Reversed"
                elif orient_temp == "Reversed":
                    orient = "Forward"
            else:
                orient = face.Orientation

            if "Sphere" in surface_type:
                surface_type = "Sphere"

            # cone additional plane is added afterward
            if surface_type in ("<Cylinder object>", "<Cone object>", "Sphere") and orient == "Reversed":
                # cone additional plane is added afterward
                id_face = get_id(face.Surface, surfaces, options, tolerances, numeric_format)
                if surface_type == "<Cone object>":
                    cones.add(id_face)
                if str(id_face) not in surf_piece:
                    surf_piece.append(str(id_face))
                    surf_obj.append(face)

                try:
                    plane = gen_plane(face, solid_gu, tolerances)
                    if plane is not None:
                        plane = GU.PlaneGu(plane)
                except:
                    plane = None
                    logger.warning("generation of additional plane has failed")

                if plane is not None:
                    p = GeounedSurface(
                        ("Plane", (plane.Position, plane.Axis, plane.dim1, plane.dim2)),
                        universe_box,
                        Face="Build",
                    )

                    id, exist = surfaces.add_plane(p, options, tolerances, numeric_format, False)
                    sign = sign_plane(face.CenterOfMass, p)
                    if exist:
                        pp = surfaces.get_surface(id)
                        if is_opposite(p.Surf.Axis, pp.Surf.Axis, tolerances.angle):
                            id = -id
                    id *= sign

                    if id_face not in extra_plane_reverse.keys():
                        extra_plane_reverse[id_face] = [str(id)]
                        surf_obj.append(p.shape)
                    else:
                        if str(id) not in extra_plane_reverse[id_face]:
                            extra_plane_reverse[id_face].append(str(id))
                            surf_obj.append(p.shape)

            elif surface_type == "<Toroid object>":

                if (
                    is_parallel(face.Surface.Axis, FreeCAD.Vector(1, 0, 0), tolerances.angle)
                    or is_parallel(face.Surface.Axis, FreeCAD.Vector(0, 1, 0), tolerances.angle)
                    or is_parallel(face.Surface.Axis, FreeCAD.Vector(0, 0, 1), tolerances.angle)
                ):

                    idT = get_id(face.Surface, surfaces, options, tolerances, numeric_format)

                    index, u_params = solid_gu.TorusUParams[iface]
                    if index == last_torus:
                        continue
                    last_torus = index

                    # add if necesary additional planes following U variable
                    u_closed, u_minMax = u_params
                    # u_closed = True
                    if not u_closed:
                        planes, ORop = gen_torus_annex_u_planes(face, u_minMax, tolerances)
                        plane1, plane2 = planes
                        plane = GeounedSurface(("Plane", plane1), universe_box, Face="Build")
                        id1, exist = surfaces.add_plane(plane, options, tolerances, numeric_format, False)
                        if exist:
                            p = surfaces.get_surface(id1)
                            if is_opposite(plane.Surf.Axis, p.Surf.Axis, tolerances.pln_angle):
                                id1 = -id1

                        if plane2 is None:
                            u_var = "%i" % id1
                        else:
                            plane = GeounedSurface(("Plane", plane2), universe_box, Face="Build")
                            id2, exist = surfaces.add_plane(plane, options, tolerances, numeric_format, False)
                            if exist:
                                p = surfaces.get_surface(id2)
                                if is_opposite(plane.Surf.Axis, p.Surf.Axis, tolerances.pln_angle):
                                    id2 = -id2

                            u_var = "(%i : %i)" % (id1, id2) if ORop else "%i %i" % (id1, id2)

                    else:
                        u_var = ""

                    # add if necesary additional surface following V variable
                    if orient == "Forward":
                        v_var = "-%i" % idT

                    else:
                        index, Vparams = solid_gu.TorusVParams[iface]
                        VClosed, VminMax = Vparams
                        if VClosed:
                            v_var = "%i" % idT
                        else:
                            surf_params, surf_type, in_surf = gen_torus_annex_v_surface(
                                face, VminMax, tolerances, options.forceCylinder
                            )

                            if surf_type == "Cone":
                                cone = GeounedSurface(("Cone", surf_params), universe_box, Face="Build")
                                id2, exist = surfaces.add_cone(cone, tolerances)

                            elif surf_type == "Cylinder":
                                cyl = GeounedSurface(
                                    ("Cylinder", surf_params),
                                    universe_box,
                                    Face="Build",
                                )
                                id2, exist = surfaces.add_cylinder(cyl, options, tolerances, numeric_format)

                            elif surf_type == "Plane":
                                plane = GeounedSurface(("Plane", surf_params), universe_box, Face="Build")
                                id2, exist = surfaces.add_plane(plane, options, tolerances, numeric_format, False)
                                if exist:
                                    p = surfaces.get_surface(id2)
                                    if is_opposite(
                                        plane.Surf.Axis,
                                        p.Surf.Axis,
                                        tolerances.pln_angle,
                                    ):
                                        id2 = -id2

                            v_var = "%i %i" % (idT, -id2 if in_surf else id2)

                    var = v_var if u_closed else " ".join((v_var, u_var))
                    if var not in surf_piece:
                        surf_piece.append(var)
                        surf_obj.append(face)
                else:
                    logger.info("Only Torus with axis along X, Y, Z axis can be reproduced")
            else:
                id = get_id(face.Surface, surfaces, options, tolerances, numeric_format)
                if surface_type == "<Cone object>":
                    cones.add(-id)

                surf = face
                if id == 0:
                    logger.warning(f"{surface_type} not found in surface list")
                    if surface_type == "<Plane object>":
                        dim1 = face.ParameterRange[1] - face.ParameterRange[0]
                        dim2 = face.ParameterRange[3] - face.ParameterRange[2]
                        plane = GeounedSurface(
                            (
                                "Plane",
                                (face.Surface.Position, face.Surface.Axis, dim1, dim2),
                            ),
                            universe_box,
                            Face="Build",
                        )
                        id, exist = surfaces.add_plane(plane, options, tolerances, numeric_format, False)
                        surf = plane.shape
                    elif surface_type == "<Cylinder object>":
                        dim_l = face.ParameterRange[3] - face.ParameterRange[2]
                        cylinder = GeounedSurface(
                            (
                                "Cylinder",
                                (
                                    face.Surface.Center,
                                    face.Surface.Axis,
                                    face.Surface.Radius,
                                    dim_l,
                                ),
                            ),
                            universe_box,
                            Face="Build",
                        )
                        id, exist = surfaces.add_cylinder(cylinder, options, tolerances, numeric_format)
                        surf = cylinder.shape

                if orient == "Reversed":
                    var = id
                elif orient == "Forward":
                    var = -id

                if surface_type == "<Plane object>":
                    s = surfaces.get_surface(id)
                    if is_opposite(face.Surface.Axis, s.Surf.Axis, tolerances.pln_angle):
                        var = -var

                if str(var) in surf_piece:
                    continue

                surf_piece.append(str(var))
                surf_obj.append(surf)

        if extra_plane_reverse:
            for extra in extra_plane_reverse.values():
                if len(extra) == 1:
                    if extra[0] not in surf_piece:
                        surf_piece.append(extra[0])
                else:
                    surf_piece.append(f"({':'.join(extra)})")

        surf_piece_bool = BoolSequence(" ".join(surf_piece))
        # possible expresion for e
        #  i1
        #  i1 i2surf_piece
        #  i1 i2 i3
        #  (i1:i2)
        #  i1 (i2:i3)

        #         semi = e.find(':')
        #         blk  = e.find(' ')
        #         print (e)
        #         if semi != -1 :
        #           orTerm = expOR(int(e[1:semi]),int(e[semi+1:-1]))
        #           surf_piece_bool.add(orTerm)
        #         elif blk != -1 :
        #           surf_piece_bool.add(int(e[1:blk]),int(e[blk+1:-1]))
        #         else :
        #          surf_piece_bool.add(int(e))

        if surf_piece_bool.elements:
            piece_def.append(surf_piece_bool)
            iece_obj.append(surf_obj)
        else:
            del_list.append(isol)

    for isol in reversed(del_list):
        del meta_obj.Solids[isol]
    meta_obj.set_definition(piece_def)
    meta_obj.set_faces(iece_obj)
    return tuple(cones)


def get_surf_value(definition, reverse=False):

    if definition.level == 0:
        if reverse:
            surf = {-i for i in definition.elements}
        else:
            surf = set(definition.elements)
    else:
        surf = set()
        for e in definition.elements:
            if e.operator == "AND":
                if reverse:
                    surf = {-i for i in e.elements}
                else:
                    surf = set(e.elements)
                break
    return surf


def append_comp(new_cell, cell_def, cell_cad, meta_complementary):

    surf_cell = get_surf_value(cell_def, True)
    if meta_complementary.Definition.operator == "AND":
        if not cell_cad.BoundBox.intersect(meta_complementary.CADSolid.BoundBox):
            return False
        Seq = meta_complementary.Definition
        surfComp = get_surf_value(Seq, False)
        if len(surfComp & surf_cell) > 0:
            return False
        new_cell.append(Seq.get_complementary())
        return True

    else:
        append = False
        for i, compPart in enumerate(meta_complementary.Solids):
            if not cell_cad.BoundBox.intersect(compPart.BoundBox):
                continue
            Seq = meta_complementary.Definition.elements[i]
            surfComp = get_surf_value(Seq, False)
            if len(surfComp & surf_cell) > 0:
                continue
            append = True
            new_cell.append(Seq.get_complementary())
        return append


def no_overlapping_cell(metaList, surfaces, options):

    Surfs = {}
    for lst in surfaces.values():
        for s in lst:
            Surfs[s.Index] = s

    new_definition_list = []
    metaList[0].set_cad_solid()

    for i, m in enumerate(metaList[1:]):
        m.set_cad_solid()

        if m.Definition.operator == "AND":
            new_def = BoolSequence(operator="AND")
            new_def.append(m.Definition.copy())
            simplify = False
            for mm in metaList[: i + 1]:
                simp = append_comp(new_def, m.Definition, m.CADSolid, mm)
                if simp:
                    simplify = True
            simpTerm = [simplify]

        else:
            new_def = BoolSequence(operator="OR")
            simpTerm = []
            for j, partSolid in enumerate(m.Solids):
                subDef = BoolSequence(operator="AND")
                subDef.append(m.Definition.elements[j].copy())
                simplify = False
                for mm in metaList[: i + 1]:
                    simp = append_comp(subDef, m.Definition.elements[j], partSolid, mm)
                    if simp:
                        simplify = True
                simpTerm.append(simplify)
                new_def.append(subDef)
        new_definition_list.append((new_def, simpTerm))

    for m, t_def_and_simplify in zip(metaList[1:], new_definition_list):
        t_def, simplify = t_def_and_simplify
        if True in simplify:
            logger.info(f"reduce cell {m.__id__}")
            box = UF.get_box(m, options.enlargeBox)

            # evaluate only diagonal elements of the Constraint Table (fastest) and remove surface not
            # crossing in the solid boundBox
            CT = build_c_table_from_solids(
                box,
                (tuple(t_def.get_surfaces_numbers()), Surfs),
                "diag",
                options=options,
            )

            new_def = remove_extra_surfaces(t_def, CT)

            # evaluate full constraint Table with less surfaces involved
            CT = build_c_table_from_solids(
                box,
                (tuple(new_def.get_surfaces_numbers()), Surfs),
                "full",
                options=options,
            )

            if new_def.operator == "AND":
                new_def.simplify(CT)
                new_def.clean()
            else:
                for i, s in enumerate(simplify):
                    if not s:
                        continue
                    comp = new_def.elements[i]
                    comp.simplify(CT)
                    comp.clean()

            m.set_definition(new_def)
            m.Definition.join_operators()
            m.Definition.level_update()


def add_cone_plane(definition, cones_list, surfaces, universe_box, options, tolerances, numeric_format):
    x_axis = FreeCAD.Vector(1, 0, 0)
    y_axis = FreeCAD.Vector(0, 1, 0)
    z_axis = FreeCAD.Vector(0, 0, 1)

    for cid in cones_list:
        cone = surfaces.get_surface(abs(cid))
        if (
            is_parallel(cone.Surf.Axis, x_axis, tolerances.angle)
            or is_parallel(cone.Surf.Axis, y_axis, tolerances.angle)
            or is_parallel(cone.Surf.Axis, z_axis, tolerances.angle)
        ):
            continue

        plane = GeounedSurface(
            ("Plane", (cone.Surf.Apex, cone.Surf.Axis, 1, 1)),
            universe_box,
            Face="Build",
        )
        pid, exist = surfaces.add_plane(plane, options, tolerances, numeric_format, False)

        if exist:
            p = surfaces.get_surface(pid)
            if is_opposite(plane.Surf.Axis, p.Surf.Axis, tolerances.pln_angle):
                pid = -pid

        if cid > 0:
            insert_in_sequence(definition, cid, -pid, "OR")
        else:
            insert_in_sequence(definition, cid, pid, "AND")



# -*- coding: utf-8 -*-

"""
Functions for parsing MCNP input files.
"""

import re
import warnings

from .PartialFormatter import PartialFormatter

# integer with one prefix character
re_int = re.compile(r"\D{0,1}\d+")

# interior of square brackets for tally in lattices
re_ind = re.compile(r"\[.+\]", flags=re.DOTALL)

# repitition syntax of MCNP input file
re_rpt = re.compile(r"\d+[ri]", flags=re.IGNORECASE)

# imp or tmp parameters in cell card
re_prm = re.compile(r"((imp:n|imp:p|tmp)\s+\S+)")
re_prm = re.compile(r"[it]mp:*[npe]*[=\s]+\S+", flags=re.IGNORECASE)
re_prm = re.compile(r"([it]mp:*[npe]*[=\s]+)(\S+)", flags=re.IGNORECASE)

# fill keyword
re_fll = re.compile(r"\*{0,1}fill[=\s]+", flags=re.IGNORECASE)  # TODO: this will also match fill===


# If type specifier not given, any data type can be formatted:
def fmt_gen(s):
    return "{" + f":<{len(s)}" + "}"


fmt_d = fmt_gen
fmt_g = fmt_gen
fmt_s = fmt_gen

partial_formmatter = PartialFormatter()


class CidClass(object):
    """
    There are two levels of card types. 1-st level is purely defined by card
    position in the input file.  There can be:
          * message card
          * title card
          * cell card
          * surface card
          * data card.

    Data cards can be of different type that is defined by the card's first
    entry. Therefore data cards can be characterized by type of the 2-nd level:
          * m cards,
          * f cards,
          * etc.

    This class is to describe the 1-st level of card types.
    """

    # no-information cards
    comment = -1
    blankline = -2
    # card types in order they appear in MCNP input
    message = 1
    title = 2
    cell = 3
    surface = 4
    data = 5

    @classmethod
    def get_name(cls, cid):
        """
        Return the name of the card type by its index.
        """
        for k, v in list(cls.__dict__.items()):
            if "__" not in k and v == cid:
                return k
        else:
            print("No attribute with name", cid)
            raise ValueError()


CID = CidClass()


class Card(object):
    """
    Representation of a card.
    """

    def __init__(self, lines, ctype, pos, debug=None):

        # Original lines, as read from the input file
        self.lines = lines

        # card type by its position in the input. See CID class.
        self.ctype = ctype

        # True if self.lines has changed after initialization
        # used in remove_hash function
        self.cstrg = False

        # data card type. Defined from the get_values() method.
        # Has sense only to data cards (see ctype). For other card types
        # is None.
        self.dtype = None

        # Input file line number, where the card was found.
        self.pos = pos

        # File-like object to write debug info (if not None)
        self.debug = debug

        # template string. Represents the general structure of the card. It is
        # a copy of lines, but meaningful parts are replaced by format
        # specifiers, {}
        self.template = ""

        # List of strings represenging meaningful parts of the card. The
        # original multi-line string card is obtained as
        # template.format(*input)
        self.input = []

        # Dictionary of parts that are removed from input before processing it.
        # For example, repitition syntax (e.g. 5r or 7i) is replaced with '!'
        # to prevent its modification.
        self.hidden = {}

        # List of (v, t) tuples, where v -- value and t -- its type.
        self.values = []

        # geometry prefix and suffix
        # self.geom_prefix = ''
        # self.geom_suffix = ''

        # some properties defined on demand
        # cell properties
        self.__u = -1  # -1 means undefined. None -- not specified in input
        self.__f = -1  # fill
        self.__m = -1  # material
        self.__d = ""  # density
        self.__i = -1  # importances
        self.__cr = -1  # set of reference cells.
        # surface properties
        self.__st = ""  # '' means undefined.

        # Split card to template and meaningful part is always needed. Other
        # operations are optional.
        self.get_input()
        return

    def _get_value_by_type(self, t):
        """
        Returns the first value of type t found in self.values.
        """
        vl, tl = zip(*self.values)
        try:
            i = tl.index(t)
        except ValueError:
            return None
        finally:
            return vl[i]

    def _set_value_by_type(self, t, v):
        """
        Sets the first value of type t to v in self.values.
        """
        vl, tl = zip(*self.values)
        i = tl.index(t)
        self.values[i] = (v, t)

    @property
    def geom_prefix(self):
        return self._get_value_by_type("#gpr")

    @geom_prefix.setter
    def geom_prefix(self, value):
        return self._set_value_by_type("#gpr", value)

    @property
    def geom_suffix(self):
        return self._get_value_by_type("#gsu")

    @geom_suffix.setter
    def geom_suffix(self, value):
        return self._set_value_by_type("#gsu", value)

    def print_debug(self, comment, key="tihv"):
        d = self.debug
        if d:
            print(
                f"Line {self.pos}, {CID.get_name(self.ctype)} card. {comment}",
                file=d,
            )
            if "t" in key:
                print("    template:", repr(self.template), file=d)
            if "i" in key:
                print("    input:   ", self.input, file=d)
            if "h" in key:
                print("    hidden:  ", self.hidden, file=d)
            if "v" in key:
                print("    values:  ", self.values, file=d)

    def get_input(self, check_bad_chars=False):
        """
        Recompute template, input and hidden attributes from lines
        """

        mline = "".join(self.lines)
        if check_bad_chars:
            bad_chars = "\t"
            for c in bad_chars:
                if c in mline:
                    if self.debug:
                        self.print_debug("get_input: bad char in input cards", "")
                    else:
                        raise ValueError("Bad character in input file. " + "Run with --debug option.")

        if self.ctype in (CID.comment, CID.blankline):
            # nothing to do for comments or blanklines:
            self.input = ""
            self.template = mline

        else:
            # TODO: protect { and } in comment parts of the card.
            tmpl = []  # part of template
            inpt = []  # input, meaningful parts of the card.
            if mline.split()[0][:2] == "fc":
                # this is tally comment. It always in one line and is not
                # delimited by & or $
                i = mline[:80]
                t = mline.replace(i, "{}", 1)
                inpt = [i]
                tmpl = [t]
            else:
                for l in self.lines:
                    if is_commented(l):
                        tmpl.append(l)
                    else:
                        # entries optionally delimited from comments by $ or &
                        # requires that delimiters prefixed with space
                        d = index_(l, "$&")
                        # d1 = l.find(' $')
                        # d2 = l.find(' &')
                        # if -1 < d1 and -1 < d2:
                        #     # both & and $ in the line. Use the most left
                        #     d = min(d1, d2)
                        # elif d1 == -1 and d2 == -1:
                        #     # no delimiters at all, whole line is meaningful
                        #     # except the new-line char
                        #     d = len(l) - 1
                        # else:
                        #     # only one of delimiters is given.
                        #     d = max(d1, d2)
                        i = l[:d]
                        t = l[d:]
                        inpt.append(i)
                        tmpl.append(fmt_s(i) + t)
            self.input = inpt
            self.template = "".join(tmpl)

            # TODO: dtype and name of the card can be defined already here.

        self.print_debug("get_input", "ti")
        return

    def _protect_nums(self):
        """
        In the meaningful part of the card replace numbers that do not
        represent cell, surface or a cell parameter with some unused char.
        """

        inpt = "\n".join(self.input)

        d = {}

        # in cell card:
        if self.ctype == CID.cell:  # and 'like' not in inpt:
            d["~"] = []  # float values in cells

            # Replace material density
            if "like" not in inpt:
                tokens = inpt.replace("=", " ").split()
                cell, mat, rho = tokens[:3]
                if int(mat) != 0:
                    for s in (cell, mat, rho):
                        inpt = inpt.replace(s, "~", 1)
                    inpt = inpt.replace("~", cell, 1)
                    inpt = inpt.replace("~", mat, 1)
                    d["~"].append(rho)

            # imp and tmp parameters:
            # print 're_prm: inp', repr(inpt)
            for s1, s2 in re_prm.findall(inpt):
                # print 're_prm: fnd', repr(s)
                d["~"].append(s2)
                inpt = inpt.replace(s1 + s2, s1 + "~", 1)

        # replace repitition syntax in junks:
        sbl = re_rpt.findall(inpt)
        if sbl:
            for s in sbl:
                inpt = inpt.replace(s, "!", 1)
            d["!"] = sbl

        if self.ctype == CID.data and inpt.lstrip().lower()[0] == "f" and inpt.lstrip()[1].isdigit():
            # this is tally card. Hide indexes in square brackets
            sbl = re_ind.findall(inpt)
            if sbl:
                for s in sbl:
                    inpt = inpt.replace(s, "|", 1)
                d["|"] = sbl

        self.input = inpt.split("\n")
        self.hidden = d

        self.print_debug("_protect_nums", "ih")
        return

    def get_values(self):
        """
        Replace integers in the meaningfull part with format specifiers, and
        populate the `values` attribute.
        """
        self._protect_nums()
        if self.ctype == CID.cell:
            inpt, vt = _split_cell(self.input, self)
            self.name = vt[0][0]
        elif self.ctype == CID.surface:
            inpt, vt, stype, scoef = _split_surface(self.input)
            self.TR = None
            if len(vt) > 1:
                if vt[1][1] == "tr":
                    self.TR = vt[1][0]
            self.stype = stype.lower()
            self.scoefs = scoef
            self.name = vt[0][0]
        elif self.ctype == CID.data:
            inpt, vt, dtype = _split_data(self.input)
            self.dtype = dtype
            if dtype == "TRn":
                # print(vt,inpt)
                unit, inpt, fvals = _parse_tr(inpt)
                self.unit = unit
                vt += fvals
            if self.dtype is not None:
                self.name = vt[0][0]
        else:
            inpt = self.input
            vt = []

        self.input = inpt
        self.values = vt

        self.print_debug("get_values", "iv")
        return

    def get_refcells(self):
        """
        Returns all cells used in definition of self.
        """
        if self.ctype != CID.cell:
            return None
        if self.__cr != -1:
            return self.__cr
        else:
            s = set()
            for v, t in self.values:
                if t == "cel":
                    s.add(v)
            self.__cr = s
            return self.__cr

    def get_geom(self):
        """
        Returns part of the cell card describing geometry, as a (multiline)
        string.
        """
        p, s = self.geom_prefix, self.geom_suffix
        self.geom_prefix = "§"
        self.geom_suffix = "§"
        geom = self.card().split("§")[1]
        self.geom_prefix = p
        self.geom_suffix = s
        return geom

    def get_u(self):
        """
        Returns universe, the cells belongs to.
        """
        if self.ctype != CID.cell:
            return None
        if self.__u != -1:
            return self.__u
        else:
            # get it only once:
            for v, t in self.values:
                if t == "u":
                    self.__u = v
                    break
            else:
                self.__u = None
            return self.__u

    def get_m(self):
        """
        For cell card return material number
        """
        if self.ctype != CID.cell:
            return None

        if self.__m != -1:
            return self.__m
        else:
            if "like" in "".join(self.input).lower():
                # material name should be given in another cell.
                pass
            for v, t in self.values:
                if t == "mat":
                    self.__m = v
                    break
            else:
                # raise ValueError("Cell does not have material specs")
                self.__m = -2
            return self.__m

    def get_d(self):
        """
        For cell card return density
        """
        if self.__d != "":
            return self.__d

        if self.get_m() == 0:
            self.__d = 0
            return self.__d
        elif self.get_m() == -2:
            # this is like-but cell
            self.__d = -100.0
            return self.__d
        else:
            # density entry is hidden in the input and available as the 1-st
            # entry in self.hidden dictionary.
            self.__d = float(self.hidden["~"][0])
            return self.__d

    def set_d(self, v):
        """
        Set density. Accespted are string represetntaion of a float.

        It is assumed that get_values() method is called before this.
        """
        if self.get_m() > 0:
            self.hidden["~"][0] = v
            self.__d = float(v)

    def get_f(self, newv=None, trsf=False):
        """
        Returns universe, the cell is filled with.
        """
        if self.ctype != CID.cell:
            return None

        if self.__f != -1 and newv is None:
            if trsf:
                params = []
                for v, t in self.values:
                    if t == "#tunit":
                        unit = v
                    elif t == "tr":
                        params = [v]
                        break
                    elif t == "#tparam":
                        params.append(float(v))
                return self.__f, (unit, params)
            else:
                return self.__f
        else:
            # get it only once:
            unit = ""
            params = []
            for i in range(len(self.values)):
                v, t = self.values[i]
                if t == "fill":
                    if newv is not None:
                        v = newv
                        self.values[i] = (v, t)
                    self.__f = v
                    if trsf:
                        unit = self.values[i - 1][0]
                        for v, t in self.values[i + 1 :]:
                            if t == "tr":
                                params = [v]
                                break
                            elif t == "#tparam":
                                params.append(float(v))
                    break
            else:
                self.__f = None

            if trsf:
                return self.__f, (unit, params)
            else:
                return self.__f

    def get_imp(self, vals={}):
        """
        Returns importances, if explicitly specified in the cell card.
        """
        if self.ctype != CID.cell:
            return None

        if self.__i != -1 and not vals:
            return self.__i
        else:
            res = {}
            inpt = " ".join(self.input).lower()
            for p in "npe":
                key = "imp:" + p

                s = inpt.split(key)
                if len(s) == 1:
                    # there is no key in the input line.
                    continue
                else:
                    n = s[0].count("~")
                    res[key] = float(self.hidden["~"][n])
                    if p in vals:
                        # change value only if necessary
                        if res[key] != vals[p]:
                            res[key] = vals[p]
                            self.hidden["~"][n] = str(vals[p])

                # for s in self.hidden.get('~', []):
                #     sl = s.lower()
                #     if key in sl:
                #         val = float(sl.replace(key, '').replace('=', ''))
                #         res[key] = val
            if not res:
                res["imp:n"] = 1
            self.__i = res
            return self.__i

    def remove_fill(self):
        """
        Removes the FILL= keyword of a cell card.

        This method must be called after get_values().
        """
        # Fill card is followed by one universe number and optionally by
        # transformation in parentheses. Optionally, the `fill` keyword can be
        # prefixed with asterisk

        # Two possibilites are here: (1) start from original line and clean out
        # everything related to FILL, or (2) modify already existing template,
        # values and input. In both cases, the new card should be checked for
        # empty lines.

        # Case (2): Modify existing input and values. All content after the
        # FILL keyword is parsed and thus is given in values, while input
        # provides placefor them. THus, simply replaceing values with spaces
        # will almost do the job. The remaining part -- the keyword itself that
        # is presented in input.

        # replace with spaces all FILL-related tokens
        vals = []  # new values list.
        oldv = self.values[:]
        state = "before"
        while oldv:
            v, t = oldv.pop(0)
            if state == "before" and t == "fill":
                v = " "
                state = "afterU"
            elif state == "afterU" and "(" in t:
                v = " "
                state = "after("
            elif state == "after(":
                v = " "
                if ")" in t:
                    state = "after"
            vals.append((v, t))
        self.values = vals

        # Remove FILL from the input
        for n, i in enumerate(self.input):
            if "fill" in i.lower():
                # This part of input contains the fill keyword. This keyword is
                # optionally prepended with an asterix and followed by a sign
                i = re_fll.sub(" ", i)
                self.input[n] = i
                break

        self.print_debug("remove_fill", "iv")
        return

    def card(self, wrap=False, comment=True):
        """
        Return multi-line string representing the card.
        """
        if self.input:
            # put values back to meaningful parts:
            inpt = "\n".join(self.input)
            inpt = inpt.format(*[t[0] for t in self.values])

            # put back hidden parts:
            for k, vl in list(self.hidden.items()):
                for v in vl:
                    inpt = inpt.replace(k, v, 1)

            inpt = inpt.split("\n")
            if not comment:
                return " ".join(inpt)

            if wrap:  # and self.ctype != CID.title:
                indent = " " * 5
                if self.ctype == CID.title:
                    indent = "c" + indent
                tparts = re.split(r"\{.*?\}", self.template)[1:]
                # print 'wrapped inp', repr(self.template)
                # print 'wrapped spl', repr(tparts)
                newt = [""]  # new template parts
                newi = []  # new input parts
                self.print_debug("card wrap=True", "")
                for i, t in zip(inpt, tparts):
                    self.print_debug("    " + repr(i) + repr(t), "")
                    il = []
                    tl = [t]

                    # while len(i.rstrip()) > 79:
                    while len(i.rstrip()) > 80:
                        # first try to shift to left
                        if i[:5] == " " * 5:
                            i = " " * 5 + i.lstrip()
                        if len(i.rstrip()) > 79:
                            # input i must be wrapped. Find proper place:
                            for dc in " :":
                                k = i.rstrip().rfind(dc, 0, 75)
                                if k > 6:
                                    il.append(i[:k])
                                    tl.append("\n")
                                    i = indent + i[k:]
                                    self.print_debug("card wrap=True" + repr(il[-1]) + repr(i), "")
                                    break
                            else:
                                # there is no proper place to wrap.
                                self.print_debug("Cannot wrap line " + repr(i), "")
                                warnings.warn(f"Cannot wrap card on line {self.pos}")
                                break
                        else:
                            # input i fits to one line. Do nothing.
                            pass

                    newt += tl
                    newi += il + [i]
                tmpl = "{}".join(newt)
                inpt = newi
            else:
                tmpl = self.template

            card = partial_formmatter.format(tmpl, *inpt)
            # card = tmpl.format(*inpt)
        else:
            card = self.template
        return card

    def remove_spaces(self):
        """
        Remove extra spaces from meaningful parts.
        """
        self.print_debug("before remove_spaces", "i")
        if self.ctype in (CID.cell, CID.surface, CID.data):
            inpt = []
            for i in self.input:
                indented = i[:5] == " " * 5
                # leave only one sep. space
                i = " ".join(i.split())
                i = i.strip()
                # spaces before/after some characters are not needed:
                for c in "):":
                    i = i.replace(" " + c, c)
                for c in "(:":
                    i = i.replace(c + " ", c)
                if indented:
                    i = " " * 5 + i
                inpt.append(i)
                self.print_debug(i, "")
            self.input = inpt
            self.print_debug("after remove_spaces", "i")
        return

    def apply_map(self, f):
        """
        Replace Ni in self.values by Mi = f(Ni, Ti).
        """
        self.print_debug("before apply_map", "vi")

        # u and fill should be renumberd in the same way, but types
        # must remain different, to let explicit u=0
        # self.values = map(lambda t: (f(t[0], t[1]), t[1]), self.values)
        newvals = []
        for t in self.values:
            if t[1] == "fill":
                t1 = "u"
            else:
                t1 = t[1]
            # newvals.append((f(t[0], t1), t[1]))
            if t1 in f:
                newval = f[t1](t[0])
            else:
                newval = t[0]
            newvals.append((newval, t[1]))
        self.values = newvals
        self.print_debug("after apply_map", "vi")
        return


# def _parse_geom(geom):
#     """
#     Parse the geometry part of a cell card.
#     """
#     raise NotImplementedError()
#     t = geom.split()
#     vals = []
#     fmts = []
#
#     # cell name
#     js = t.pop(0)
#     geom = geom.replace(js, tp, 1)
#     vals.append((int(js), 'cel'))
#     fmts.append(fmt_d(js))
#
#     if 'like' in geom.lower():
#         # this is like-but syntax
#         pass
#     else:
#         # get material and density.
#         # Density, if specified in cells card, should be allready hidden
#         ms = t.pop(0)
#         if int(ms) == 0:
#             inpt = inpt.replace(ms, tp+tp , 1)
#         else:
#             inpt = inpt.replace(ms, tp, 1)
#             inpt = inpt.replace('~', '~'+tp, 1)
#         vals.append((int(ms), 'mat'))
#         fmts.append(fmt_d(ms))
#
#         # placeholder for geometry prefix
#         vals.append(('', '#gpr'))
#         fmts.append('{}')


def _split_cell(input_, self):
    """
    Replace integers in the meaningful parts of a cell card with format
    specifiers, and return a list of replaced values together with their types.

    """

    # meaningful parts together. Originally, they cannot have \n chars, since
    # all of them should land to the card template, therefore, after all
    # entries are replaced with format specifiers, it can be split back to a
    # list easily at \n positions.
    inpt = "\n".join(input_)

    vals = []  # list of values
    fmts = []  # value format. It has digits, thus inserted into inpt later.
    tp = "_"  # temporary placeholder for format specifiers

    # Parse part before parameters. This is different for usual and like-but
    # syntax.  As result, all entries are replaced in inpt and i, the index
    # where parameter's part starts in inpt, is computed.

    if "like " in inpt.lower():

        # Get cell name
        t = inpt.split()
        js = t.pop(0)
        inpt = inpt.replace(js, tp, 1)
        vals.append((int(js), "cel"))
        fmts.append(fmt_d(js))

        # Get reference cell name:
        t.pop(0)  # like
        js = t.pop(0)
        inpt = inpt.replace(js, tp, 1)
        vals.append((int(js), "cel"))
        fmts.append(fmt_d(js))

        # compute i -- where first param token starts
        t.pop(0)  # but
        p0 = t.pop(0)
        i = inpt.index(p0)
        parm = [p0] + t

    else:
        # cell card has usual format.

        t = inpt.split()

        # Get cell name
        js = t.pop(0)
        inpt = inpt.replace(js, tp, 1)
        vals.append((int(js), "cel"))
        fmts.append(fmt_d(js))

        # get material and density.
        # Density, if specified in cells card, should be allready hidden
        ms = t.pop(0)
        if int(ms) == 0:
            inpt = inpt.replace(ms, tp + tp, 1)
        else:
            inpt = inpt.replace(ms, tp, 1)
            inpt = inpt.replace("~", "~" + tp, 1)
        vals.append((int(ms), "mat"))
        fmts.append(fmt_d(ms))

        # placeholder for geometry prefix
        vals.append(("", "#gpr"))
        fmts.append("{}")

        # Get geometry and parameters blocks. I assume that geom and param
        # blocks are separated by at least one space, so there will be an
        # element in t starting with alpha char -- This will be the first token
        # from the param block.
        geom = []
        parm = []
        while t:
            e = t.pop(0)
            if e[0].isalpha() or e[0] == "*":
                parm = [e] + t
                break
            else:
                geom.append(e)

        # print '_split_cell geom', geom, parm
        # replace integer entries in geom block:
        for s in re_int.findall(" ".join(geom)):
            # print 's from re_int', repr(s)
            # s is a surface or a cell (later only if prefixed by #)
            t = "cel" if s[0] == "#" else "sur"
            s = s if s[0].isdigit() else s[1:]
            f = fmt_d(s)
            inpt = inpt.replace(s, tp, 1)
            # print 't', repr(t)
            # print 's', repr(s)
            # print 'f', repr(f)
            # print repr(inpt)
            vals.append((int(s), t))
            fmts.append(f)

        # geometry suffix
        vals.append(("", "#gsu"))
        fmts.append("{}")
        # insert placeholder for geometry suffix
        if parm:
            inpt = inpt.replace(parm[0], "_" + parm[0], 1)

        # At this point all geom entries are replaced in inpt. The rest should
        # work only with the parm part of inpt. To ensure this, inpt is splitted
        # into inpt_geom and inpt_parm:
        if parm:
            i = inpt.index(parm[0])
        else:
            i = len(inpt)

    inpt_geom = inpt[:i]
    inpt_parm = inpt[i:]

    # print 'From parsing'
    # print repr(inpt)
    # print i
    # print repr(inpt_geom)
    # print repr(inpt_parm)
    # print parm

    # replace values in parameters block. Values are prefixed with = or space(s)
    # Note that tmp and imp values must be hidden
    t = " ".join(parm).replace("=", " ").split()  # get rid of =.
    while t:
        s = t.pop(0)
        # print '_split_cell s: ', repr(s)
        if s.lower() == "u":
            vs = t.pop(0)
            vv = int(vs)
            vf = fmt_d(vs)
            vt = "u"
            inpt_parm = inpt_parm.replace(vs, tp, 1)
            vals.append((vv, vt))
            fmts.append(vf)
        elif "fill" in s.lower():
            # print '_split_cell: has fill!'
            # assume that only one integer follows the fill keyword, optionally
            # with transformation in parentheses.

            # set transformation dimension deg if *FILL or cos if FILL
            if "*" in s:
                vv = "*"
            else:
                vv = ""
            vf = fmt_s(vv)
            vt = "#tunit"
            vals.append((vv, vt))
            fmts.append(vf)

            vs = t.pop(0)
            # if transformation in parentheses follows the universe number
            # immediately, split this manually:
            if "(" in vs:
                i = vs.index("(")
                ttt = vs[i:]
                vs = vs[:i]
                # vs, ttt = vs.split('(')
                t.insert(0, ttt)
            vv = int(vs)
            vf = fmt_d(vs)
            vt = "fill"
            inpt_parm = inpt_parm.replace(vs, tp, 1)
            vals.append((vv, vt))
            fmts.append(vf)
            # fill value can be followed by transformation in parentheses
            # Fill value can be optionally followed by transformation number of
            # transformation parameters in parentheses
            if t and "(" in t[0]:
                vsl = []  # lists of strings, values, formats and types
                vvl = []
                vfl = []
                vtl = []

                # add opening parenthesis
                vsl.append("(")
                vvl.append("(")
                vfl.append(fmt_s("("))
                vtl.append("#(")  # #-types are internal, don't output in --mode info.
                t[0] = t[0].replace("(", "", 1)

                # add entries in parentheses and the closing parenthis
                while vsl[-1] != ")":
                    vs = t.pop(0)
                    if ")" in vs:
                        vs = vs.replace(")", "", 1)
                        if vs:
                            vsl.append(vs)
                            vvl.append(vs)
                            vfl.append(fmt_s(vs))
                            vtl.append("#tparam")
                        vsl.append(")")
                        vvl.append(")")
                        vfl.append(fmt_s(")"))
                        vtl.append("#)")
                    elif vs:
                        vsl.append(vs)
                        vvl.append(vs)
                        vfl.append(fmt_s(vs))
                        vtl.append("#tparam")

                # check if only one parameter in parenthethes -- it is tr
                # number, not tr parameter
                if len(vsl) == 3:
                    vvl[1] = int(vvl[1])
                    vfl[1] = fmt_d(vsl[1])
                    vtl[1] = "tr"

                # add all strings, values, formats and types:
                for vs, vv, vf, vt in zip(vsl, vvl, vfl, vtl):
                    inpt_parm = inpt_parm.replace(vs, tp, 1)  # TODO: here only parm part of inpt should be modified.
                    vals.append((vv, vt))
                    fmts.append(vf)

            # warn if there is possibility for an array following the fill
            # keyword:
            # TODO fill value can be an array
            if "fill" == s.lower() and "lat" in "".join(parm).lower():
                print("WARNING: fill keyword followed by an array", end=" ")
                print("cannot be parsed")

    inpt = inpt_geom + inpt_parm

    # replace '_' with fmts:
    for f in fmts:
        inpt = inpt.replace(tp, f, 1)

    return inpt.split("\n"), vals


def _split_surface(input_):
    """
    Similar to _split_cell(), but for surface cards.
    """
    inpt = "\n".join(input_)
    t = inpt.split()

    vals = []  # like in split_cell()
    fmts = []
    tp = "_"

    # get surface name:
    js = t.pop(0)
    if not js[0].isdigit():
        js = js[1:]
    inpt = inpt.replace(js, tp, 1)
    vals.append((int(js), "sur"))
    fmts.append(fmt_d(js))

    # get TR or periodic surface:
    ns = t.pop(0)
    if ns[0].isdigit():
        # TR is given
        inpt = inpt.replace(ns, tp, 1)
        vals.append((int(ns), "tr"))
        fmts.append(fmt_d(ns))
        st = t.pop(0)
    elif ns[0] == "-":
        # periodic surface
        ns = ns[1:]
        inpt = inpt.replace(ns, tp, 1)
        vals.append((int(ns), "sur"))
        fmts.append(fmt_d(ns))
        st = t.pop(0)
    elif ns[0].isalpha():
        # ns is the surface type
        st = ns
    else:
        raise ValueError(input_, inpt, ns)

    # define coefficients
    scoef = list(map(float, t))

    for f in fmts:
        inpt = inpt.replace(tp, f, 1)

    return inpt.split("\n"), vals, st, scoef


def _get_int(s):
    r = ""
    for c in s:
        if r and c.isalpha():
            break
        elif c.isdigit():
            r += c
    return r


def _parse_tr(input_):
    """
    input_ should be already passed through _split_data()
    """
    inpt = "\n".join(input_)
    inp1, inp2 = inpt.split(None, 1)
    if "*" in inp1:
        unit = "*"
    else:
        unit = ""

    svals = inp2.split()
    for s in svals:
        inp2 = inp2.replace(s, "{}", 1)

    fvals = [(float(s), "float") for s in svals]
    return unit, (inp1 + " " + inp2).split("\n"), fvals


def _split_data(input_):
    inpt = "\n".join(input_)
    t = inpt.split()

    vals = []
    fmts = []
    tp = "_"

    if "tr" in t[0][:3].lower():
        # TRn card
        dtype = "TRn"
        ns = _get_int(t[0])
        inpt = inpt.replace(ns, tp, 1)
        vals.append((int(ns), "tr"))
        fmts.append(fmt_d(ns))
    elif t[0][0].lower() == "m" and "mode" not in t[0].lower() and "mesh" not in t[0].lower() and "mphys" not in t[0].lower():
        # This is the Mn, MTn or MPNn card
        ms = _get_int(t[0])
        inpt = inpt.replace(ms, tp, 1)
        vals.append((int(ms), "mat"))
        fmts.append(fmt_d(ms))
        # additional tests to define data card type:
        if t[0][1].isdigit():
            dtype = "Mn"
        elif t[0][1].lower() == "t":
            dtype = "MTn"
        elif t[0][1].lower() == "p":
            dtype = "MPNn"
        elif t[0][1].lower() == "x":
            dtype = "MXn"
        else:
            dtype = None
    elif t[0][0].lower() == "f" and t[0][1].isdigit():
        # FN card
        dtype = "Fn"
        ns = _get_int(t[0])  # tally number
        inpt = inpt.replace(ns, tp, 1)
        vals.append((int(ns), "tal"))
        fmts.append(fmt_d(ns))

        # define type of integers by tally type:
        nv = int(ns[-1])
        if nv in [1, 2]:
            typ = "sur"
        elif nv in [4, 6, 7, 8]:
            typ = "cel"
        else:
            typ = ""

        if typ:
            # Lattice indices, surrounded by square brakets must allready be
            # hidden

            # Special treatment, if tally has 'u=' syntax.
            hasu = "u" in inpt.lower() and "=" in inpt.lower()
            # find all integers -- they are cells or surfaces
            for s in re_int.findall(inpt):
                ss = s[1:]
                tpe = typ
                if hasu:
                    # ss can be universe. To distinguish this, one needs to look
                    # back in previous cheracters in c.
                    i1 = inpt.rfind(tp)
                    i2 = inpt.find(ss)
                    part = inpt[i1:i2]
                    while " " in part:
                        part = part.replace(" ", "")
                    if part[-2:].lower() == "u=":
                        tpe = "u"
                inpt = inpt.replace(ss, tp, 1)
                vals.append((int(ss), tpe))
                fmts.append(fmt_d(ss))
    elif "fmesh" == t[0][:5].lower() and t[0][5].isdigit():
        # fmesh card
        dtype = "fmesh"
        ns = _get_int(t[0])  # tally number
        inpt = inpt.replace(ns, tp, 1)
        vals.append((int(ns), "tal"))
        fmts.append(fmt_d(ns))
    else:
        dtype = None

    for f in fmts:
        inpt = inpt.replace(tp, f, 1)

    return inpt.split("\n"), vals, dtype


def is_commented(l):
    """
    Return True if l is a commented line.
    """
    res = False

    # remove newline chars at the end of l:
    l = l.splitlines()[0]
    ls = l[0:6].lstrip().lower()
    if "c " == ls[0:2]:
        res = True
        # print 'is_com "c "',
    elif "c" == l.lower():
        res = True
        # print 'is_com "c"',
    # print 'is_com', res
    return res


def is_fc_card(l):
    """
    Return true, if line l is tally comment cards, fcN
    """
    return l.lstrip().lower()[:2] == "fc"


def is_blankline(l):
    """
    Return True, if l is the delimiter blank line.
    """
    return l.strip() == ""


def get_cards(inp, debug=None, preservetabs=False):
    """
    Check first existence of a dump file

    If dump exists and it is newwer than the input file, read the dump file
    """
    for c in get_cards_from_input(inp, debug=debug, preservetabs=preservetabs):
        yield c


def index_(line, chars="$&"):
    """
    Find the first index of one of the chars in line.
    """
    r = re.compile(f"[{chars}]")
    m = r.search(line)
    if m:
        i = m.end() - 1
    else:
        i = len(line) - 1
    return i


def get_cards_from_input(inp, debug=None, preservetabs=False):
    """
    Iterable, return instances of the Card() class representing
    cards in the input file.

    inp -- is the filename.
    """

    def _yield(card, ct, ln):
        return Card(card, ct, ln, debug)

    def replace_tab(l, cln, preserve=False, ts=8):
        """
        Replace tabs as in MCNP5 (Vol II, Chapter 1 - Primer, I. MCNP INPUT FOR
        SAMPLE PROBLEM, A. INP File, p. 1-3)
        """
        if preserve:
            return l[:]
        else:
            while "\t" in l:
                i = l.index("\t")
                ii = (i // ts + 1) * ts - i
                print(f"c Line {cln + 1}: tab replaced with {ii} spaces")
                l = l[:i] + " " * ii + l[i + 1 :]
            return l[:]

    cln = 0  # current line number. Used only for debug
    with open(inp, "r") as f:
        # define the first block:
        # -----------------------

        # Next block ID
        ncid = 0  # 0 is not used in card ID dictionary CID.

        # Parse the 1-st line. It can be message, cell or data block.
        l = replace_tab(next(f), cln, preserve=preservetabs)
        cln += 1
        # kw = l.lower().split()[0]
        kw = l.lstrip()
        if "message:" == kw[:8].lower():
            # read message block right here
            res = []
            while not is_blankline(l):
                res.append(l)
                l = replace_tab(next(f), cln, preserve=preservetabs)
                cln += 1
            yield _yield(res, CID.message, cln - 1)  # message card
            yield _yield(l, CID.blankline, cln)  # blank line
            l = replace_tab(next(f), cln, preserve=preservetabs)
            cln += 1
            ncid = CID.title
        elif "continue" == kw[:8].lower():
            # input file for continue job. Contains only data block.
            ncid = CID.data
        else:
            ncid = CID.title
        if ncid == CID.title:
            # l contains the title card
            yield _yield([l], ncid, cln)
            ncid += 1

        # read all other lines
        # --------------------

        # Line can be a continuation line in the following cases:
        #   * all lines in the message block, i.e. before the blank line
        #     delimiter
        #   * if line starts with 5 or more spaces,
        #   * if previous line ends with & sign.
        # Thus, the role of the current line (continuation or not) can be
        # defined by the line itself (5 spaces), or by previous lines (message
        # block or & sign). This can lead to inconsistency, when previous line
        # is delimited by &, but is followed by the blank line delimiter.  in
        # this case (rather theretical), blank line delimiter (as appeared more
        # lately) delimites the card from the previous line.
        cf = False  # continuation line flag. True only when prev. line has &.

        # Comment lines (CL) can be between cards or inside them. CL between two
        # cards are yielded as block of comments (although usually, CL are used
        # to describe cards that follow them).  CL inside a card will belong to
        # the card.

        card = []  # card is a list of lines.
        cmnt = []  # list of comment lines.
        for l in f:
            l = replace_tab(l, cln, preserve=preservetabs)
            cln += 1
            if is_blankline(l):
                # blank line delimiter. Stops card even if previous line
                # contains &
                if card:
                    # card can be empty, for example, when several empty lines
                    # are at the end of file
                    yield _yield(card, ncid, cln - len(card) - len(cmnt))
                if cmnt:
                    yield _yield(cmnt, CID.comment, cln - len(cmnt))
                    cmnt = []
                yield _yield(l, CID.blankline, cln)
                ncid += 1
                card = []
                if ncid == 6:
                    break
            elif l[0:5] == "     " or cf:
                # l is continuation line.
                if cmnt:
                    card += cmnt  # prev. comment lines belong to this card.
                    cmnt = []
                card.append(l)
                cf = l[: index_(l)].find("&", 0, 81) > -1
            elif is_commented(l):
                # l is a line comment. Where it belongs (to the current card or
                # to the next one), depends on the next line, therefore, just
                # store temorarily.
                cmnt.append(l)
            else:
                # l is the 1-st line of a card. Return previous card and
                # comments
                if card:
                    yield _yield(card, ncid, cln - len(card) - len(cmnt))
                if cmnt:
                    yield _yield(cmnt, CID.comment, cln - len(cmnt))
                    cmnt = []
                card = [l]
                # if tally comment card, i.e. started with fc, the & character
                # does not mean continuation.
                cf = not is_fc_card(l) and l[: index_(l)].find("&", 0, 81) > -1
        if card:
            yield _yield(card, ncid, cln - len(card) - len(cmnt))
        if cmnt:
            yield _yield(cmnt, CID.comment, cln - len(cmnt))


def get_blocks(cards):
    """
    Return a dict of cards in blocks.
    """

    d = {}
    cbt = None  # current block type
    cbc = []  # current block cards
    for c in cards:
        if c.ctype == CID.blankline:
            d[cbt] = cbc
            cbt = None
            cbc = []
        elif c.ctype == CID.title:
            d[c.ctype] = [c]
        else:
            cbc.append(c)
            if cbt is None and c.ctype > 0:
                cbt = c.ctype
    if cbc:
        d[cbt] = cbc
    return d


def are_close_vals(x, y, re=1e-6, ra=0.0):
    """
    Return True if x and y are closer then re or ra.
    """
    if abs(x - y) <= ra:
        r = True
    elif x != 0:
        r = abs((x - y) / x) <= re
    else:
        # y is not equal to x and x is 0 -> y is not 0.
        r = abs((x - y) / y) <= re
    return r


def are_close_lists(x, y, re=1e-6, pci=[]):
    """
    Return True if x and y are close but not equal.
    """
    if len(x) != len(y):
        res = False
        msg = "Different length"

    if x == y:
        return True

    # pci -- list of indices that define elements of x and y to be checked for
    # proportionality only.
    if len(pci) == 0:
        # empty list means all x and y elements compare without arbitrary
        # normalization.
        xe = x[:]
        ye = y[:]
        xp = []
        yp = []
    else:
        if len(pci) % 2 == 1:
            # augment with len(x) +1
            pci = tuple(pci) + (len(x) + 1,)
        xe = []
        ye = []
        xp = []
        yp = []
        i = 0
        for i1, i2 in zip(pci[0::2], pci[1::2]):
            xe += x[i:i1]
            ye += y[i:i1]
            xp += x[i1:i2]
            yp += y[i1:i2]
            i = i2

    # normalize yp
    xpn = sum([e**2 for e in xp])
    ypn = sum([e**2 for e in yp])
    if xpn > 0 and ypn > 0:
        yp = [e * xpn / ypn for e in yp]

    msg = []
    res = []
    for xl, yl in zip([xe, xp], [ye, yp]):
        # compare xl and yl without normalization
        if xl == yl:
            res.append(True)
            msg.append("exact match")
        else:
            n = 0
            for xx, yy in zip(xl, yl):
                r = are_close_vals(xx, yy, re)
                if not r:
                    m = f"diff at {n}"
                    break
            else:
                m = "all elements are close or equal"
                r = True
            res.append(r)
            msg.append(m)

        if not res[-1]:
            result = False
            break

    else:
        result = True
    return result


if __name__ == "__main__":
    pass



"""
The tests check the resulting volume of the CSG cells using OpenMC.
The tests assumes that the test_convert script has previously been run and
that xml files exist in the tests_outputs folder.
"""

import math
import os
import sys
from pathlib import Path

try:
    import freecad  # importing conda package if present
except:
    pass
import geouned
import openmc
import Part
import pytest
from FreeCAD import Import

with open("cross_sections.xml", "w") as file:
    file.write(
        """
        <?xml version='1.0' encoding='UTF-8'?>
        <cross_sections>
        <library materials="H1" path="" type="neutron"/>
        </cross_sections>
    """
    )
openmc.config["cross_sections"] = Path("cross_sections.xml").resolve()

path_to_cad = Path("testing/inputSTEP")
step_files = list(path_to_cad.rglob("*.stp")) + list(path_to_cad.rglob("*.step"))
# this file is removed as it fails the test. An issue has been raised to track
step_files.remove(Path("testing/inputSTEP/Misc/rails.stp"))
# this checks if the tests are being run a github action runner or not
if os.getenv("GITHUB_ACTIONS"):
    # reduced samples as github action runners have 2 threads and more samples
    # is not needed for the smaller models tested. This allows the CI to
    # quickly test the smaller models
    samples = 4_000_000
    rel_tol = 0.05
    # removing two geometries that are particularly slow to convert from CI testing
    # these two geometries remain in the test suite for locally testing
    step_files.remove(Path("testing/inputSTEP/large/SCDR.stp"))
    step_files.remove(Path("testing/inputSTEP/large/Triangle.stp"))
else:
    # samples for local run can be larger as threads is likely to be larger
    samples = 40_000_000
    # acceptable tolerance can also be smaller
    rel_tol = 0.04


@pytest.mark.skipif(
    sys.platform in ["win32", "darwin"],
    reason="OpenMC doesn't install on Windows currently and is not well tested on Mac",
)
@pytest.mark.parametrize("input_step_file", step_files)
def test_volumes(input_step_file):

    output_dir = Path("tests_outputs") / input_step_file.with_suffix("")
    output_dir.mkdir(parents=True, exist_ok=True)
    output_filename_stem = output_dir / input_step_file.stem
    openmc_xml_file = output_filename_stem.with_suffix(".xml")

    Import.insert(str(input_step_file), "converted-cad")
    result = Part.Shape()
    result.read(str(input_step_file))

    materials = openmc.Materials()
    geometry = openmc.Geometry.from_xml(openmc_xml_file, materials)
    cells = geometry.get_all_cells()

    # some of the csg cells are void space around the CAD, so we just check
    # that there are at least as many cells as there could be more cells in the
    # CSG than the CAD
    assert len(result.Solids) <= len(cells.keys())

    all_cell_vol_calc = []
    for solid, (cell_id, cell) in zip(result.Solids, cells.items()):

        bb = solid.BoundBox
        llx, lly, llz, urx, ury, urz = (
            bb.XMin,
            bb.YMin,
            bb.ZMin,
            bb.XMax,
            bb.YMax,
            bb.ZMax,
        )
        llx, lly, llz, urx, ury, urz = (
            llx / 10,
            lly / 10,
            llz / 10,
            urx / 10,
            ury / 10,
            urz / 10,
        )

        cell_vol_calc = openmc.VolumeCalculation(
            domains=[cell],
            samples=samples,
            lower_left=(llx, lly, llz),
            upper_right=(urx, ury, urz),
        )
        # could add a trigger to ends tests early once they have converged
        # cell_vol_calc.set_trigger(0.05, 'rel_err')

        all_cell_vol_calc.append(cell_vol_calc)

    settings = openmc.Settings()
    settings.volume_calculations = all_cell_vol_calc
    settings.run_mode = "volume"
    model = openmc.Model(geometry, materials, settings)

    model.export_to_model_xml(path=openmc_xml_file.with_name("model.xml"))
    openmc.run(cwd=output_dir)

    for solid, (cell_id, cell) in zip(result.Solids, cells.items()):

        vol_calc_file = openmc_xml_file.with_name(f"volume_{cell_id}.h5")
        cell_vol_calc_results = openmc.VolumeCalculation.from_hdf5(vol_calc_file)
        volume_of_csg_cell = cell_vol_calc_results.volumes[cell_id].nominal_value
        # converts from mm3 in cad to cm3 in csg
        volume_of_cad_cell = solid.Volume * 0.001
        assert math.isclose(volume_of_cad_cell, volume_of_csg_cell, rel_tol=rel_tol)


@pytest.mark.parametrize("skip_solids,expected", [([], 12), ([1, 4, 6], 9)])
def test_skip_solids_when_loading(skip_solids, expected):
    """test to check that the correct number of cells are found in the CSG file when skip solids is set"""

    geo = geouned.CadToCsg(settings=geouned.Settings(voidGen=False, compSolids=False))

    # this geometry has 12 solids in it
    geo.load_step_file(filename="testing/inputSTEP/Torus/example.stp", skip_solids=skip_solids)
    geo.start()
    assert geo.filename == "testing/inputSTEP/Torus/example.stp"
    assert geo.skip_solids == skip_solids

    geo.export_csg(
        geometryName="tests_outputs/skip_solids/csg",
        outFormat=["openmc_xml"],
    )

    materials = openmc.Materials()
    geometry = openmc.Geometry.from_xml("tests_outputs/skip_solids/csg.xml", materials)
    cells = geometry.get_all_cells()
    all_cell_ids = cells.keys()
    assert len(all_cell_ids) == expected



import typing
import FreeCAD
import Import

from pathlib import Path

from .Modules.buildCAD import buildCAD, makeTree
from .Modules.MCNPinput import McnpInput
from .Modules.Objects import CadCell
from .Modules.XMLinput import XmlInput


class CsgToCad:
    """Base class for the conversion of CSG to CAD models"""

    def __init__(self):
        pass

    def export_cad(
        self,
        input_filename: str,
        csg_format: str,
        output_filename: str = "cad_from_csg",
        bounding_box: typing.Tuple[int, int, int, int, int, int] = (-1000, -1000, -1000, 1000, 1000, 1000),
        universe_start: int = 0,
        level_max: str = "all",
        cell_range_type: str = "all",
        cell_range: typing.Tuple[int] = (),
        mat_range_type: str = "all",
        mat_range: typing.Tuple[int] = (),
        # TODO add these to the method signature
        # splitTolerance in the Options class
        # mat =  this is in the CADselection dictionary but not in the docs https://github.com/GEOUNED-org/GEOUNED/blob/76ef697c7dca6a828c7498996ff3313859c872f2/docs/User_Guide_GEOUNED_v2.0.pdf
    ):
        """export the CSG geometry in OpenMC or MCNP format to a CAD model.

        Args:
            input_filename (str): The filename and path of the input CSG text file.
            csg_format (str): The format of the CSG input file, options are 'openmc_xml' or 'mcnp'
            output_filename (str, optional): The filename stem and path of the output file created.
                Two files will be created with the '.step' suffix and one with the 'FCStd' suffix.
                Defaults to 'cad_from_csg'.
            bounding_box (typing.Tuple[int, int, int, int, int, int], optional): The bounding box
                coordinates of the CSG geometry. This should encompass the entire CSG geometry.
                Format is (xmin, ymin, zmin, xmax, ymax, zmax) Defaults to (-1000, -1000, -1000,
                1000, 1000, 1000).
            universe_start (int, optional): The Universe ID to be converted to CAD. If universe_start
                is left as 0 then all the universes and any nested universes are converted. If set
                then the universe and all its nested universes are converted. Defaults to 0.
            level_max (str, optional): Level maximum of nested universe to be translated. If
                level_max < highest nested universe level, cells inside the container cell whose
                level is level_max will not be translated. This container cell will be the CAD
                solid written in the CAD file. Defaults to "all".
            cell_range_type (str, optional): Define how to consider the range values.
                setting to 'all' results in all the cells with any cell ID being converted (range a no effect).
                setting to 'include' results in only cells defined in 'range' being converted.
                setting to 'exclude' results in exclude all cells defined in range. Defaults to "all".
            cell_range (typing.Tuple[int], optional): list of cells to be included/excluded for the conversion.
                Defaults to ().
            mat_range_type (str, optional): Define how to consider the range values.
                setting to 'all' results in all the materials with any cell ID being converted (range a no effect).
                setting to 'include' results in only materials defined in 'range' being converted.
                setting to 'exclude' results in exclude all materials defined in range. Defaults to "all".
            mat_range (typing.Tuple[int], optional): list of materials to be included/excluded for the conversion.
                Defaults to ().

        Raises:
            ValueError: If the csg_format is not 'openmc_xml' or 'mcnp' then a ValueError is raised.
        """

        # TODO check file extensions are correct
        # if Path(output_filename).suffix not in ['.stp', '.step']:
        #     raise ValueError(f"output file must have a .stp or .step extension, not {universe_start.suffix}")

        # get geometry definition from OpenMC XML or MCNP input
        if csg_format == "mcnp":
            geo = McnpInput(input_filename)
        elif csg_format == "openmc_xml":
            geo = XmlInput(input_filename)
        else:
            msg = f"input format type {csg_format} is not supported. Supported options are 'openmc_xml' or 'mcnp'"
            raise ValueError(msg)

        Path(output_filename).parent.mkdir(parents=True, exist_ok=True)

        UnivCell = CadCell()
        UnivCell.shape = UnivCell.makeBox(FreeCAD.BoundBox(*bounding_box))

        # TODO make these variable names lower case in the downstream code

        CADselection = {
            "Ustart": universe_start,
            "levelMax": level_max,
            "cell": [cell_range_type, cell_range],
            "mat": [mat_range_type, mat_range],
            "format": csg_format,
            "cell_range_type": cell_range_type,
            "cell_range": cell_range,
            "mat_range_type": mat_range_type,
            "mat_range": mat_range,
        }

        # TODO don't return fails variable, just fail in the method and raise the error there
        CADCells, fails = buildCAD(UnivCell, geo, CADselection)

        if fails:
            print("failed in conversion", fails)

        CADdoc = FreeCAD.newDocument("converted_with_geouned")

        makeTree(CADdoc, CADCells)
        Import.export(CADdoc.Objects[0:1], f"{output_filename}.step")
        CADdoc.saveAs(f"{output_filename}.FCStd")



import geouned


def test_version():
    """Check that __version__ exists and is correctly formatted"""
    version = geouned.__version__
    # Ensure it is given as a string
    assert isinstance(version, str)

    # develop is used as the default version name in the dockerfile
    if version != "develop":
        version_bits = version.split(".")
        assert len(version_bits) >= 2, f"Version number is {version}"
        # Ensure both of the first two parts is convertable to int
        # (The third/fourth part, if it exists, may be a development version)
        for bit in version_bits[:2]:
            assert bit.isdigit(), f"Version number is {version}"
